# -*- coding: UTF-8 -*-
# @Time    : 18-4-27
# @File    : demo.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.ant.base import *
from antgo.ant.base import _pick_idle_port
from antgo.dataflow.dataset.queue_dataset import *
from antgo.dataflow.recorder import *
from antgo.crowdsource.demo_server import *
from antgo.utils import logger
from antvis.client.httprpc import *
try:
    import queue
except:
    import Queue as queue
import traceback

class AntDemo(AntBase):
  def __init__(self, ant_context,
               ant_name,
               ant_dump_dir,
               ant_token,
               ant_task_config,
               **kwargs):
    super(AntDemo, self).__init__(ant_name, ant_context, ant_token, **kwargs)

    self.ant_dump_dir = ant_dump_dir
    self.ant_context.ant = self
    self.ant_task_config = ant_task_config

    self.html_template = kwargs.get('html_template', None)
    self.demo_port = kwargs.get('port', None)
    self.demo_port = int(self.demo_port) if self.demo_port is not None else None
    # self.support_user_upload = kwargs.get('support_user_upload', False)
    # self.support_user_input = kwargs.get('support_user_input', False)
    # self.support_user_interaction = kwargs.get('support_user_interaction', False)
    # self.support_user_constraint = kwargs.get('support_user_constraint', None)
    self.context.devices = [int(d) for d in kwargs.get('devices', '').split(',') if d != '']

  def start(self):
    # 1.step loading demo task
    running_ant_task = None
    if self.token is not None:
      # 1.1.step load challenge task
      response = mlogger.getEnv().dashboard.challenge.get(command=type(self).__name__)
      if response['status'] == 'ERROR':
        # invalid token
        logger.error('couldnt load challenge task')
        self.token = None
      elif response['status'] == 'SUSPEND':
        # prohibit submit challenge task frequently
        # submit only one in one week
        logger.error('prohibit submit challenge task frequently')
        exit(-1)
      elif response['status'] == 'OK':
        content = response['content']
        # maybe user token or task token
        if 'task' in content:
          challenge_task = create_task_from_json(content)
          if challenge_task is None:
            logger.error('couldnt load challenge task')
            exit(-1)
          running_ant_task = challenge_task
      else:
        # unknow error
        logger.error('unknow error')
        exit(-1)

    if running_ant_task is None:
      # 1.2.step load custom task
      custom_task = create_task_from_xml(self.ant_task_config, self.context)
      if custom_task is None:
        logger.error('couldnt load custom task')
        exit(-1)
      running_ant_task = custom_task

    assert (running_ant_task is not None)

    # 2.step 注册实验
    experiment_uuid = self.context.experiment_uuid
    # 3.step 配置数据传输管道
    dataset_queue = queue.Queue()

    demo_dataset = QueueDataset(dataset_queue)
    demo_dataset._force_inputs_dirty()

    # 3.step 配置dump路径
    infer_dump_dir = os.path.join(self.ant_dump_dir, experiment_uuid, 'inference')
    if not os.path.exists(infer_dump_dir):
      os.makedirs(infer_dump_dir)
    else:
      shutil.rmtree(infer_dump_dir)
      os.makedirs(infer_dump_dir)

    # 4.step 启动web服务
    # 4.1.step 选择合适端口
    if self.demo_port is None:
      self.demo_port = 10000
    self.demo_port = _pick_idle_port(self.demo_port)
    logger.info('Demo prepare using port %d.'%self.demo_port)

    # 后台api调用
    self.rpc = HttpRpc("v1", "api", "127.0.0.1", self.demo_port)

    demo_name = running_ant_task.task_name
    demo_type = running_ant_task.task_type
    demo_config = {
      'interaction':{
        'support_user_upload': True,
        'support_user_input': True,
        'support_user_interaction': False,
        'support_user_constraint': {}
      },
      'description_config': self.description_config,
      'port': self.demo_port,
      'html_template': self.html_template,
      'dump_dir': infer_dump_dir
    }

    request_waiting_time = 30
    if self.context.params.demo is not None:
      if 'support_user_upload' in self.context.params.demo:
        demo_config['interaction']['support_user_upload'] = self.context.params.demo['support_user_upload']

      if 'support_user_input' in self.context.params.demo:
        demo_config['interaction']['support_user_input'] = self.context.params.demo['support_user_input']

      if 'support_user_interaction' in self.context.params.demo:
        demo_config['interaction']['support_user_interaction'] = self.context.params.demo['support_user_interaction']

      if 'support_user_constraint' in self.context.params.demo:
        demo_config['interaction']['support_user_constraint'] = self.context.params.demo['support_user_constraint']

      if 'request_waiting_time' in self.context.params.demo:
        request_waiting_time = self.context.params.demo['request_waiting_time']

    # 5.step 启动后台线程运行预测过程
    def _run_infer_process():
      # prepare ablation blocks
      logger.info('Prepare model ablation blocks.')
      ablation_blocks = getattr(self.ant_context.params, 'ablation', [])
      if ablation_blocks is None:
        ablation_blocks = []
      for b in ablation_blocks:
        self.ant_context.deactivate_block(b)

      # infer
      logger.info('Running inference process.')
      try:
        def _callback_func(data):
            id = demo_dataset.now()['id'] if 'id' in demo_dataset.now() else ''
            record_content = {
                        'experiment_uuid': experiment_uuid, 
                        'data': data,
                        'id': id
                    }
            for data_group in record_content['data']:
              for item in data_group:
                  if item['type'] in ['IMAGE', 'VIDEO', 'FILE']:
                      item['data'] = '%s/record/%s'%(infer_dump_dir, item['data'])
                
            self.rpc.response.post(response=json.dumps(record_content))
                    
        self.context.recorder =  LocalRecorderNodeV2(_callback_func)
        self.context.call_infer_process(demo_dataset, dump_dir=infer_dump_dir)
      except Exception as e:
        print(e)
        traceback.print_exc()

    process = threading.Thread(target=_run_infer_process)
    process.daemon = True
    process.start()

    # 6.step 启动web服务
    demo_server_start(demo_name,demo_type,demo_config,os.getpid(),dataset_queue, request_waiting_time)



# encoding=utf-8
# @Time    : 17-3-3
# @File    : common.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os
import tensorflow as tf
from tensorflow.python.tools import freeze_graph
from tensorflow.python.tools import optimize_for_inference_lib
from tensorflow.python.platform import gfile
slim = tf.contrib.slim
from antgo.dataflow.dataset import *
import scipy.misc


def _int64_feature(values):
  if not isinstance(values, (tuple, list)):
    values = [values]
  return tf.train.Feature(int64_list=tf.train.Int64List(value=values))


def _bytes_feature(values):
  return tf.train.Feature(bytes_list=tf.train.BytesList(value=[values]))


def _get_dataset_record_filepath(dataset_record_dir, dataset_name, split_name, shard_id, num_shards):
  output_filename = '%s_%s_%05d-of-%05d.tfrecord' % (dataset_name, split_name, shard_id, num_shards)
  return os.path.join(dataset_record_dir, output_filename)


class AntTFRecords(object):
    def __init__(self, ant_context, data_factory, dataset, dump_dir):
        self.num_shards = 100
        self.dump_dir = dump_dir
        self.ant_context = ant_context
        self.data_factory = data_factory
        self.dataset = dataset

    def __make_records(self, data):
        reorganized_data = {}
        for key, value in data.items():
            xxyy = key.split('_')
            data_name = xxyy[0]
            if data_name not in reorganized_data:
                reorganized_data[data_name] = {}

            if len(xxyy) == 1:
                reorganized_data[data_name]['data'] = value
            elif xxyy[1] == 'TYPE':
                reorganized_data[data_name]['type'] = value

        tf_data = {}
        for key, value in reorganized_data.items():
            v_data = value['data']
            v_type = value['type']
            if v_type == "IMAGE":
                # 保存临时文件，使用tf.gfile.FastGFile
                scipy.misc.imsave("./temp.png", v_data)
                v_data = tf.gfile.FastGFile('./temp.png', 'rb').read()
                tf_data[key] = _bytes_feature(v_data)
            elif v_type == "STRING":
                tf_data[key] = _bytes_feature(v_data.encode('utf-8'))
            elif v_type == "INT":
                tf_data[key] = _int64_feature(v_data)

        return tf_data

    def start(self):
        parse_flag = ''
        dataset_cls = AntDatasetFactory.dataset(self.dataset, parse_flag)

        if dataset_cls is None:
            logger.error('couldnt find dataset parse class')
            return

        # 4.step 启动数据生成
        # 4.1.step 训练集
        train_dataset = dataset_cls('train', os.path.join(self.data_factory, self.dataset))

        if train_dataset.size > 0:
            # 确保每个shard中至少100个样本
            num_shards = self.num_shards
            num_per_shard = int(np.math.ceil(train_dataset.size / float(num_shards)))
            if num_per_shard < 200:
                num_shards = int(np.math.ceil(train_dataset.size / 200))
                num_per_shard = int(np.math.ceil(train_dataset.size / num_shards))

            logger.info('train dataset -> tfrecords (%d shards, %d samples in a shard)'%(num_shards, num_per_shard))
            sample_index = 0
            try:
                tfrecord_writer = None
                with tf.Session() as sess:
                    record_dir = os.path.join(self.dump_dir, 'train')
                    if not os.path.exists(record_dir):
                        os.makedirs(record_dir)

                    for data in self.ant_context.data_generator(train_dataset):
                        if sample_index % num_per_shard == 0:
                            if tfrecord_writer is not None:
                                tfrecord_writer.close()
                                tfrecord_writer = None

                            record_filepath = \
                                _get_dataset_record_filepath(record_dir,
                                                             self.dataset,
                                                             'train',
                                                             sample_index//num_per_shard,
                                                             num_shards)

                            tfrecord_writer = tf.python_io.TFRecordWriter(record_filepath)
                            logger.info('generate train shard %d'%(sample_index//num_per_shard))

                        sample_index += 1

                        # tf features
                        feature = self.__make_records(data)
                        # warp to tf example
                        example = tf.train.Example(features=tf.train.Features(feature=feature))
                        # write to file
                        tfrecord_writer.write(example.SerializeToString())
            except:
                if tfrecord_writer is not None:
                    tfrecord_writer.close()
                    tfrecord_writer = None

                logger.info('finish training %d samples'%sample_index)

        # 4.2.step 验证集
        val_dataset = dataset_cls('val', os.path.join(self.data_factory, self.dataset))

        if val_dataset.size > 0:
            # 确保每个shard中至少100个样本
            num_shards = self.num_shards
            num_per_shard = int(np.math.ceil(val_dataset.size / float(num_shards)))
            if num_per_shard < 200:
                num_shards = int(np.math.ceil(val_dataset.size / 200))
                num_per_shard = int(np.math.ceil(val_dataset.size / num_shards))

            logger.info('val dataset -> tfrecords (%d shards, %d samples in a shard)'%(num_shards, num_per_shard))
            sample_index = 0
            try:
                tfrecord_writer = None
                with tf.Session() as sess:
                    record_dir = os.path.join(self.dump_dir, 'val')
                    if not os.path.exists(record_dir):
                        os.makedirs(record_dir)

                    for data in self.ant_context.data_generator(val_dataset):
                        if sample_index % num_per_shard == 0:
                            if tfrecord_writer is not None:
                                tfrecord_writer.close()
                                tfrecord_writer = None

                            record_filepath = \
                                _get_dataset_record_filepath(record_dir,
                                                             self.dataset,
                                                             'val',
                                                             sample_index//num_per_shard,
                                                             num_shards)

                            tfrecord_writer = tf.python_io.TFRecordWriter(record_filepath)
                            logger.info('generate val shard %d' % (sample_index // num_per_shard))

                        sample_index += 1

                        # tf features
                        feature = self.__make_records(data)
                        # warp to tf example
                        example = tf.train.Example(features=tf.train.Features(feature=feature))
                        # write to file
                        tfrecord_writer.write(example.SerializeToString())
            except:
                if tfrecord_writer is not None:
                    tfrecord_writer.close()
                    tfrecord_writer = None
                logger.info('finish val %d samples' % sample_index)

        # 4.3.step 测试集
        test_dataset = dataset_cls('test', os.path.join(self.data_factory, self.dataset))

        if test_dataset.size > 0:
            # 确保每个shard中至少100个样本
            num_shards = self.num_shards
            num_per_shard = int(np.math.ceil(test_dataset.size / float(num_shards)))
            if num_per_shard < 200:
                num_shards = int(np.math.ceil(test_dataset.size / 200))
                num_per_shard = int(np.math.ceil(test_dataset.size / num_shards))

            logger.info('test dataset -> tfrecords (%d shards, %d samples in a shard)' % (num_shards, num_per_shard))
            sample_index = 0
            try:
                tfrecord_writer = None
                with tf.Session() as sess:
                    record_dir = os.path.join(self.dump_dir, 'test')
                    if not os.path.exists(record_dir):
                        os.makedirs(record_dir)

                    for data in self.ant_context.data_generator(test_dataset):
                        if sample_index % num_per_shard == 0:
                            if tfrecord_writer is not None:
                                tfrecord_writer.close()
                                tfrecord_writer = None

                            record_filepath = \
                                _get_dataset_record_filepath(record_dir,
                                                             self.dataset,
                                                             'teset',
                                                             sample_index // num_per_shard,
                                                             num_shards)

                            tfrecord_writer = tf.python_io.TFRecordWriter(record_filepath)
                            logger.info('generate test shard %d' % (sample_index // num_per_shard))

                        sample_index += 1

                        # tf features
                        feature = self.__make_records(data)
                        # warp to tf example
                        example = tf.train.Example(features=tf.train.Features(feature=feature))
                        # write to file
                        tfrecord_writer.write(example.SerializeToString())
            except:
                if tfrecord_writer is not None:
                    tfrecord_writer.close()
                    tfrecord_writer = None

                logger.info('finish val %d samples' % sample_index)



from __future__ import unicode_literals
from __future__ import division
from antgo.dataflow.dataset import *
from antgo.measures import *
from antgo.utils import logger
from antgo import config

import json
try:
  import xml.etree.cElementTree as ET
except ImportError:
  import xml.etree.ElementTree as ET

Config = config.AntConfig
class AntTask(object):
  def __init__(self, task_id, task_name, task_type_id, task_type,
               dataset_id, dataset_name, dataset_params,
               estimation_procedure_type, estimation_procedure_params,
               evaluation_measure, cost_matrix,
               dataset_url=None,
               class_label=None,
               ext_params=None,
               ant_context=None):
    self._task_id = task_id
    self._task_name = task_name
    self._task_type_id = task_type_id
    self._task_type = task_type
    self._dataset_id = dataset_id
    self._dataset_params = dataset_params
    self._dataset_url = dataset_url
    self._estimation_procedure_type = estimation_procedure_type
    self._estimation_procedure_params = estimation_procedure_params
    self._evaluation_measure = evaluation_measure
    self._cost_matrix = cost_matrix
    self._ant_context = ant_context
    
    # config extent params
    if ext_params is not None:
      for k, v in ext_params.items():
        if k != 'self':
          setattr(self, k, v)
    
    if self._dataset_params is None:
      self._dataset_params = {}
      
    if self._dataset_url is not None and len(self._dataset_url) > 0:
      self._dataset_params['dataset_url'] = self._dataset_url

    # 
    self._dataset_name = dataset_name
    if len(dataset_name.split('/')) != 1:
      dataset_name = dataset_name.split('/')[0]
    self._dataset_params['dataset_name'] = dataset_name

    # dataset class
    if dataset_name is not None:
      parse_flag = ''
      if self._dataset_url is not None and len(self._dataset_url) > 0:
        parse_flag = self._dataset_url.split('/')[-2]
      self._ant_dataset = AntDatasetFactory.dataset(dataset_name, parse_flag)

    # related evaluation measures
    self._ant_measures = AntMeasuresFactory(self)
    self._class_label = class_label

  @property
  def dataset_name(self):
    return self._dataset_name

  @dataset_name.setter
  def dataset_name(self, name):
    self._dataset_name = name

  @property
  def dataset_params(self):
    '''

    Returns
    -------

    '''
    return self._dataset_params

  @property
  def dataset(self):
    return self._ant_dataset

  @property
  def task_name(self):
    return self._task_name

  @property
  def estimation_procedure(self):
    '''

    Returns
    -------
    return how to estimate ml model
    '''
    return self._estimation_procedure_type

  @estimation_procedure.setter
  def estimation_procedure(self, val):
    self._estimation_procedure_type = val

  @property
  def estimation_procedure_params(self):
    '''

    Returns
    -------
    return estimation procedure parameters
    '''
    return self._estimation_procedure_params

  @property
  def evaluation_measures(self):
    '''

    Returns
    -------
    return how to evaluate ml model
    '''
    return self._ant_measures.measures(self._evaluation_measure)

  def evaluation_measure(self, measure_name):
    '''

    :param measure_name:
    :return:
    '''
    return self._ant_measures.measures(measure_name)

  @property
  def cost_matrix(self):
    '''

    Returns
    -------
    return how to punish prediciton result
    '''
    return self._cost_matrix

  @property
  def class_label(self):
    return self._class_label
  @class_label.setter
  def class_label(self, val):
    self._class_label = val

  @property
  def task_id(self):
    return self._task_id

  @property
  def task_type(self):
    return self._task_type

  @staticmethod
  def support_task_types():
    return ['OBJECT-DETECTION',
            'SEGMENTATION',
            'CLASSIFICATION',
            'TRACK',
            'PERSON_SEARCH',
            'IMAGETRANSFER',
            'STYLETRANSFER',
            'IMAGE_CAPTION',
            'ACTION_CLASSIFICATION',
            'REGRESSION',
            'INSTANCE-SEGMENTATION',
            'MATTING',
            'IMAGE-GENERATION',
            'TTS',
            'LANDMARK',
            'GENERATION',
            'FACE_VERIFICATION',
            'FACE_IDENTIFICATION']


def create_dummy_task(task_type):
  return AntTask(task_id=-1,
                 task_name=None,
                 task_type_id=-1,
                 task_type=task_type,
                 dataset_id=-1,
                 dataset_name='',
                 dataset_params=None,
                 estimation_procedure_type='',
                 estimation_procedure_params=None,
                 evaluation_measure=None,
                 cost_matrix=None)


def create_task_from_json(task_config_json, ant_context=None):
  try:
    # 1.step about task basic
    task = task_config_json['task']
    task_id = task['task_id']
    task_name = task['task_name']
    task_type_id = task['task_type_id']
    task_type = task['task_type']
    task_params = task['task_params']

    # 2.step about task input
    dataset_id = -1
    dataset_name = ""
    target_feature = ""
    dataset_params = {}
    dataset_url = None
    estimation_procedure_type = ""
    inputs = task['input']
    estimation_procedure_params = {}
    task_cost_matrix = []
    task_evaluation_measures = []
    class_label = None
    task_ext_params = task_params
    for term in inputs:
      if term['name'] == 'source_data':
        dataset_id = term['data_set']['data_set_id']
        dataset_name = term['data_set']['data_set_name']
        if 'data_set_params' in term['data_set']:
          dataset_params = term['data_set']['data_set_params']
        if 'data_set_url' in term['data_set']:
          dataset_url = term['data_set']['data_set_url']
      elif term['name'] == 'estimation_procedure':
        estimation_procedure_type = term['estimation_procedure']['type']
        estimation_procedure_params = {}
        for kk in term['estimation_procedure']['parameter']:
          estimation_procedure_params[kk['name']] = kk['value']
      elif term['name'] == 'cost_matrix':
        task_cost_matrix = term['cost_matrix']
      elif term['name'] == 'evaluation_measures':
        task_evaluation_measures = []
        for measure_name, measure_param in term['evaluation_measures']['evaluation_measure']:
          task_evaluation_measures.append(measure_name)
          if measure_param is not None and len(measure_param) > 0:
            # params = {}
            for k, v in measure_param.items():
              if k is None:
                continue

              # params[k.strip()] = v
              if k.strip() not in task_ext_params:
                task_ext_params[k.strip()] = v
            # task_ext_params.update(params)
      elif term['name'] == 'info':
        if 'class_label' in term:
          class_label = term['class_label']
      elif term['name'] == 'ext':
        term.pop('name')
        task_ext_params.update(term)

    return AntTask(task_id=task_id,
                   task_name=task_name,
                   task_type_id=task_type_id,
                   task_type=task_type,
                   dataset_id=dataset_id,
                   dataset_name=dataset_name,
                   dataset_url=dataset_url,
                   dataset_params=dataset_params,
                   estimation_procedure_type=estimation_procedure_type,
                   estimation_procedure_params=estimation_procedure_params,
                   evaluation_measure=task_evaluation_measures,
                   cost_matrix=task_cost_matrix,
                   class_label=class_label,
                   ext_params=task_ext_params,
                   ant_context=ant_context)
  except:
    traceback.print_exc()
    raise sys.exc_info()[0]


def create_task_from_xml(task_config_xml, ant_context):
  try:
    task_name = None
    task_type = None
    dataset_name = None
    dataset_params = {}
    task_ext_params = {}
    estimation_procedure_type = None
    task_evaluation_measures = []
    class_label = []
    estimation_procedure_params = {}
    tree = ET.ElementTree(file=task_config_xml)
    root = tree.getroot()
    for child in root:
      if child.tag == 'task_name':
        task_name = child.text.strip()
      elif child.tag == 'task_type':
        if child.text is not None:
          task_type = child.text.strip()
      elif child.tag == "task_badcase":
        for item in child:
          if item.tag == 'badcase_num':
            task_ext_params.update({'badcase_size': (int)(item.text)})
          if item.tag == 'badcase_category':
            task_ext_params.update({'badcase_category': (int)(item.text)})
      elif child.tag == 'input':
        for input_item in child:
          if input_item.tag == 'source_data':
            for data_item in input_item:
              if data_item.tag == 'data_set_name':
                dataset_name = data_item.text.strip()
              elif data_item.tag == 'data_set_params':
                for parameter in data_item:
                  key = parameter[0].text.strip() if parameter[0].tag == 'name' else parameter[1].text.strip()
                  value = parameter[0].text.strip() if parameter[0].tag == 'value' else parameter[1].text.strip()
                  dataset_params[key] = value
          elif input_item.tag == 'estimation_procedure':
            for data_item in input_item:
              if data_item.tag == 'type':
                if data_item.text is not None:
                  estimation_procedure_type = data_item.text.strip()
              elif data_item.tag == 'parameter':
                if data_item[0].text is not None and \
                                data_item[1].text is not None:
                  key = data_item[0].text.strip() if data_item[0].tag == 'name' else data_item[1].text.strip()
                  value = data_item[0].text.strip() if data_item[0].tag == 'value' else data_item[1].text.strip()
                  estimation_procedure_params[key] = value
          elif input_item.tag == 'evaluation_measures':
            for data_item in input_item:
              if data_item.tag == 'evaluation_measure':
                if data_item.text is None:
                  continue

                task_evaluation_measures.append(data_item.text.strip())
                if len(data_item.attrib) > 0:
                  params = {}
                  for k,v in data_item.attrib.items():
                    params[data_item.text.strip()+'_'+k.strip()] = v
                  task_ext_params.update(params)
          elif input_item.tag == 'info':
            for data_item in input_item:
              if data_item.tag == 'class_label':
                for c in data_item:
                    class_label.append(c.text.strip())
          elif input_item.tag == "ext":
            pp = {}
            for data_item in input_item:
              pp[data_item.tag] = data_item.text
            if len(pp) > 0:
              task_ext_params.update(pp)

      elif child.tag == 'output':
        for input_item in child:
          if input_item.tag == 'generate_report':
            task_ext_params.update({'is_generate_report':int(input_item.text.strip())})

    return AntTask(task_id=-1,
                   task_name=task_name,
                   task_type_id=-1,
                   task_type=task_type,
                   dataset_id=-1,
                   dataset_name=dataset_name,
                   dataset_params=dataset_params,
                   estimation_procedure_type=estimation_procedure_type,
                   estimation_procedure_params=estimation_procedure_params,
                   evaluation_measure=task_evaluation_measures,
                   cost_matrix=None,
                   class_label=class_label,
                   ext_params=task_ext_params,
                   ant_context=ant_context)
  except Exception as e:
    print(e)


# -*- coding: UTF-8 -*-
# Time: 11/5/17
# File: simplecsvs.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import csv
from antgo.dataflow.dataset.dataset import *
import numpy as np
import os
import copy
import time

__all__ = ['CSV']


class CSV(Dataset):
  def __init__(self, train_or_test, dataset_dir=None, ext_params=None):
    super(CSV, self).__init__(train_or_test, dataset_dir, ext_params, 'CSV')
    dataset_path = os.path.join(dataset_dir, train_or_test)
    # make dirs
    if not os.path.exists(dataset_path):
      os.makedirs(dataset_path)

    # maybe download dataset
    all_csvs = self.all_csvs(dataset_path)
    if len(all_csvs) == 0:
      self.download(dataset_path, auto_untar=True)
      all_csvs = self.all_csvs(dataset_path)

    delimiter = getattr(self, 'delimiter', ',')
    is_skip_first = getattr(self, 'skip_first', 0)

    self.csv_data = []
    for csv_fp in all_csvs:
      with open(csv_fp,'r') as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=delimiter.encode('utf-8'))
        samples = list(csv_reader)
        if is_skip_first:
          self.csv_data.extend(samples[1:])
        else:
          self.csv_data.extend(samples)

    self.ids = np.arange(0, len(self.csv_data)).tolist()
    self.seed = time.time()

  def all_csvs(self, dataset_path):
    all_files = os.listdir(dataset_path)
    all_csvs = []
    for file_name in all_files:
      if file_name[-3:].lower() == 'csv':
        all_csvs.append(os.path.join(dataset_path, file_name))

    return all_csvs

  def data_pool(self):
    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      ids = copy.copy(self.ids)
      if self.rng:
        self.rng.shuffle(ids)

      # filter by ids
      filter_ids = getattr(self, 'filter', None)
      if filter_ids is not None:
        ids = [i for i in ids if i in filter_ids]

      # label col
      label_col = getattr(self, 'label', None)
      if label_col is not None:
        label_col = int(label_col)

      for id in ids:
        if label_col is None:
          yield [self.csv_data[id]]
        else:
          cols = len(self.csv_data[id][0])
          data = [self.csv_data[id][i] for i in range(0, cols) if i != label_col]
          label = int(self.csv_data[id][label_col])
          yield [data, label]

  def at(self, id):
    label_col = getattr(self, 'label', None)
    if label_col is not None:
      label_col = int(label_col)

    if label_col is None:
      return [self.csv_data[id]]
    else:
      cols = len(self.csv_data[id][0])
      data = [self.csv_data[id][i] for i in range(0, cols) if i != label_col]
      label = int(self.csv_data[id][label_col])
      return [data, label]

  def split(self, split_params={}, split_method="holdout"):
    assert(self.train_or_test in ['train', 'sample'])

    category_ids = copy.copy(self.ids)
    if 'is_stratified' in split_params and split_params['is_stratified'] and \
        (split_method == 'repeated-holdout' or split_method == 'holdout'):

      # label col
      label_col = getattr(self, 'label', None)
      if label_col is not None:
        label_col = int(label_col)

      if label_col is None:
        for id in self.ids:
          category_ids[id] = 0 if random.random() > 0.5 else 1
      else:
        for id in self.ids:
          category_ids[id] = int(self.csv_data[id][label_col])

    if split_method == 'kfold':
      np.random.seed(np.int64(self.seed))
      np.random.shuffle(category_ids)

    train_ids, val_ids = self._split(category_ids, split_params, split_method)
    train_dataset = CSV(self.train_or_test, self.dir, self.ext_params)
    train_dataset.ids = train_ids

    val_dataset = CSV(self.train_or_test, self.dir, self.ext_params)
    val_dataset.ids = val_ids

    return train_dataset, val_dataset

  @property
  def size(self):
    return len(self.ids)


#  -*- coding: UTF-8 -*-
# @Time    : 18-3-26
# @File    : __init__.py
# @Author  : jian<jian@mltalker.com>
from pkgutil import walk_packages
import os
import os.path
from .simplecsvs import *
from .simpleimages import *
from .simplevideos import *
from .standard import *

class AntDatasetFactory(object):
  factory_dataset = {}

  @staticmethod
  def dataset(name, parse_flag=''):
    if name in AntDatasetFactory.factory_dataset:
      return AntDatasetFactory.factory_dataset[name]

    for dataset_name, dataset_obj in AntDatasetFactory.factory_dataset.items():
      if dataset_name.lower() == name.lower():
        return dataset_obj

    if parse_flag == 'csv':
      return CSV
      
    if name.lower().startswith('image'):
      return SimpleImages
    elif name.lower().startswith('video'):
      return SimpleVideos

    return Standard

  @staticmethod
  def add_custom_dataset(custom_dataset):
    for dataset_name, dataset_obj in AntDatasetFactory.factory_dataset.items():
      if dataset_obj == custom_dataset:
        return

    AntDatasetFactory.factory_dataset[custom_dataset.__name__] = custom_dataset

def _global_import(name):
  p = __import__(name, globals(), locals(), level=1)
  globals().pop(name)
  lst = p.__all__ if '__all__' in dir(p) else dir(p)
  for k in lst:
    # add global varaible
    globals()[k] = p.__dict__[k]

    # register in Dataset Factory
    AntDatasetFactory.factory_dataset[k] = p.__dict__[k]


for _, module_name, _ in walk_packages([os.path.dirname(__file__)]):
  if not module_name.startswith('_'):
    if module_name in ['tfrecordsreader', 'dataset', 'simplecsvs', 'simpleimages', 'standard']:
      continue
    _global_import(module_name)



# -*- coding: UTF-8 -*-
# Time: 1/2/18
# File: crowdsource.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.task.task import *
from antgo.measures.base import *
from multiprocessing import Process, Queue
import numpy as np
from antgo.utils.encode import *
from antgo.utils import logger
import multiprocessing
import json
import time
import signal
import requests
from antgo.crowdsource.crowdsource_server import *


def _is_open(check_ip, port):
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  try:
    s.connect((check_ip, int(port)))
    s.shutdown(2)
    return True
  except:
    return False


def _pick_idle_port(from_port=40000, to_port=50000, check_count=20):
  check_port = None
  while check_count:
    check_port = random.randint(from_port, to_port)
    if not _is_open('127.0.0.1', check_port):
      break

    check_count = check_count - 1

  return check_port


class AntCrowdsource(AntMeasure):
  def __init__(self, task, name):
    super(AntCrowdsource, self).__init__(task, name)
    # regular crowdsource evaluation config
    self._min_participants_per_sample = getattr(task, 'min_participants_per_sample', 1)       # (from task)
    self._skip_sample_num = getattr(task, 'skip_sample_num', 2)                               # (from task)
    self._waiting_time_per_sample = getattr(task, 'waiting_time_per_sample', -1)              # unit: second (from task)
    self._max_time_in_session = getattr(task, 'max_time_in_session', -1)                      # unit: second (max time in one session) (from task)
    self._max_samples_in_session = getattr(task, 'max_samples_in_session', -1)                # max samples in one session  (from task)

    self._client_query_data = {}
    self._total_samples = 0
    self._client_html_template = ''
    self._client_keywords_template = {}

    # {CLIENT_ID, {ID: [], RESPONSE: [], RESPONSE_TIME: [], START_TIME:, QUERY_INDEX:, IS_FINISHED:}}
    self._client_response_record = {}           # client response record
    self._sample_finished_count = {}            # sample number which has been finished
    self._dump_dir = ''

    # crowdsource measure flag
    self.crowdsource = True

    self._app_token = None
    self._experiment_id = None

    self._complex_degree = 0                    # 0,1,2,3,4,5 (subclass define)
    self._crowdsource_type = ''                 # (subclass define)
    self._crowdsource_bonum = getattr(task, 'crowdsource_evaluation_bonus', 0)   # (task define)
    self._crowdsource_start_time = time.time()  # (auto)
    self._crowdsource_title = ''                # description (subclass define)
    self._crowdsource_estimated_time = 0.0      # (auto)
    self._crowdsource_complete_degree = 0       # (auto)

  @property
  def client_html_template(self):
    return self._client_html_template

  @client_html_template.setter
  def client_html_template(self, val):
    self._client_html_template = val

  @property
  def client_keywords_template(self):
    return self._client_keywords_template

  @client_keywords_template.setter
  def client_keywords_template(self, val):
    self._client_keywords_template = val

  @property
  def client_query_data(self):
    return self._client_query_data

  @client_query_data.setter
  def client_query_data(self, val):
    self._client_query_data = val
  
  @property
  def dump_dir(self):
    return self._dump_dir

  @dump_dir.setter
  def dump_dir(self, val):
    self._dump_dir = os.path.join(val,'static')

  @property
  def app_token(self):
    return self._app_token

  @app_token.setter
  def app_token(self, val):
    self._app_token = val

  @property
  def experiment_id(self):
    return self._experiment_id

  @experiment_id.setter
  def experiment_id(self, val):
    self._experiment_id = val

  @property
  def _is_finished(self):
    update_real_statistic = {}
    if self._total_samples > 0:
      # compute crowdsource complete degree
      complete_degree = 0.0
      for k,v in self._sample_finished_count.items():
        complete_degree = complete_degree + v['COUNT']
      complete_degree = complete_degree / float(self._total_samples * self._min_participants_per_sample + 1e-6)
      complete_degree = min(max(complete_degree, 0.0), 1.0)
      self._crowdsource_complete_degree = complete_degree
      update_real_statistic['complete'] = self._crowdsource_complete_degree

    # compute crodsource estimated time per client
    estimated_time = []
    for _, resposne_record in self._client_response_record.items():
      cc = [c['STOP_TIME']-c['START_TIME'] for c in resposne_record['RESPONSE_TIME'] if c is not None and c['STOP_TIME'] > 0]
      estimated_time.extend(cc)

    if len(estimated_time) > 0:
      estimated_time = np.median(estimated_time)
      self._crowdsource_estimated_time = estimated_time
      update_real_statistic['estimated_time'] = self._crowdsource_estimated_time

    # check is over?
    if len(self._sample_finished_count) == 0:
      return False

    # check is over?
    if len(self._sample_finished_count) < self._total_samples:
      return False
    
    # finish status measure - min participants per sample
    for k, v in self._sample_finished_count.items():
      if v['COUNT'] < self._min_participants_per_sample:
        return False

    return True

  def ground_truth_response(self, client_id, query_index, record_db):
    return {}

  def _prepare_ground_truth_page(self, client_id, query_index, record_db):
    response_data = {'PAGE_STATUS': 'GROUNDTRUTH', 'QUERY_INDEX': query_index}
    response_data['PAGE_DATA'] = {}

    custom_response = self.ground_truth_response(client_id, query_index, record_db)
    response_data['PAGE_DATA'].update(custom_response)
    return response_data

  def _prepare_next_page(self, client_id, query_index, record_db):
    if query_index == len(self._client_response_record[client_id]['ID']) - 1:
      return self._prepare_stop_page()

    now_time = time.time()
    query_index = query_index + 1
    self._client_response_record[client_id]['RESPONSE_TIME'][query_index] = {'START_TIME': now_time, 'STOP_TIME': -1}
    self._client_response_record[client_id]['QUERY_INDEX'] = query_index
    
    # 2.step server response
    tags = [b for a, b in self.client_keywords_template.items()]
    datas = record_db.read(self._client_response_record[client_id]['ID'][query_index], *tags)
    data_map = {}
    for a, b in zip(tags, datas):
      data_map[a] = b

    response_data = {}
    response_data['PAGE_DATA'] = {}
    # 2.1.step prepare page data
    for k, v in self.client_query_data['QUERY'].items():
      # k is name, v is type(IMAGE, SOUND, TEXT)
      data = data_map[k]

      if v == 'IMAGE':
        # save to .png
        data_en = png_encode(data)
        with open(os.path.join(self.dump_dir, '%s-%d.png' % (k,self._client_response_record[client_id]['ID'][query_index])), 'wb') as fp:
          fp.write(data_en)

        response_data['PAGE_DATA'][k] = {'DATA': '%s-%d.png' % (k,self._client_response_record[client_id]['ID'][query_index]), 'TYPE': 'IMAGE'}
        pass
      elif v == 'SOUND':
        # save
        pass
      else:
        response_data['PAGE_DATA'][k] = {'DATA': data, 'TYPE': 'TEXT'}

    # 2.3.step prepare sample index
    response_data['QUERY_INDEX'] = query_index
    # 2.4.step condition
    response_data['WAITING_TIME'] = self._waiting_time_per_sample
    # 2.5.step remained num
    response_data['REMAINED_NUM'] = len(self._client_response_record[client_id]['ID']) - query_index - 1

    return response_data

  def _prepare_stop_page(self):
    return {'PAGE_STATUS': 'STOP'}

  def _start_click_branch(self, client_query, record_db):
    client_id = client_query['CLIENT_ID']
    if client_id in self._client_response_record and self._client_response_record[client_id]['QUERY_INDEX'] != 0:
      # contiue unfinished session
      logger.info('client id %s restart unfinished session'%client_id)
      # user reenter crowdsource evaluation
      # 1.step update start time
      last_query_index = self._client_response_record[client_id]['QUERY_INDEX']
      
      last_active_time = -1
      first_sample_time = None
      for sample_response in self._client_response_record[client_id]['RESPONSE_TIME'][0:last_query_index]:
        if sample_response is not None:
          sample_start_time = sample_response['START_TIME']
          if first_sample_time is None:
            first_sample_time = sample_start_time
          
          if first_sample_time > sample_start_time:
            first_sample_time = sample_start_time
          
          sample_stop_time = sample_response['STOP_TIME']
          if sample_stop_time < -1:
            break
          
          if last_active_time < sample_stop_time:
            last_active_time = sample_stop_time

      now_time = time.time()
      if last_active_time == -1 or first_sample_time is None:
        self._client_response_record[client_id]['START_TIME'] = now_time
      else:
        self._client_response_record[client_id]['START_TIME'] = now_time - (last_active_time - first_sample_time)
      client_query['QUERY_INDEX'] = self._client_response_record[client_id]['QUERY_INDEX'] - 1
      client_query['CLIENT_RESPONSE'] = None
      # client_query['QUERY_INDEX'] may be -1
      return self._next_click_branch(client_query, record_db, True)
    else:
      # clear
      self._client_response_record[client_id] = {}
      # add client_id record
      logger.info('enter client id %s "START" response'%client_id)
      # start a new session
      # 1.step update client response_record
      ids = list(range(int(record_db.count)))
      np.random.shuffle(ids)
      client_record = {}
      # 1.1.step prepare data index order
      client_record['ID'] = ids
      client_record['RESPONSE'] = [None for _ in ids]
      # 1.2.step record start task time
      now_time = time.time()
      client_record['START_TIME'] = now_time
      # 1.3.step record first sample time
      client_record['RESPONSE_TIME'] = [None for _ in ids]
      client_record['RESPONSE_TIME'][0] = {'START_TIME': now_time, 'STOP_TIME': -1}
      client_record['QUERY_INDEX'] = 0
      self._client_response_record[client_id] = client_record

      # 2.step server response
      tags = [b for a,b in self.client_keywords_template.items()]
      datas = record_db.read(client_record['ID'][0], *tags)
      data_map = {}
      for a,b in zip(tags, datas):
        data_map[a] = b

      response_data = {}
      response_data['PAGE_STATUS'] = 'PREDICT'
      response_data['PAGE_DATA'] = {}
      # 2.1.step prepare page data
      for k, v in self.client_query_data['QUERY'].items():
        # k is name, v is type(IMAGE, SOUND, TEXT)
        data = data_map[k]

        if v == 'IMAGE':
          # save to .png
          data_en = png_encode(data)
          with open(os.path.join(self.dump_dir, '%s-%d.png'%(k, client_record['ID'][0])), 'wb') as fp:
            fp.write(data_en)

          response_data['PAGE_DATA'][k] = {'DATA': '%s-%d.png'%(k, client_record['ID'][0]), 'TYPE': 'IMAGE'}
          pass
        elif v == 'SOUND':
          # save
          pass
        else:
          response_data['PAGE_DATA'][k] = {'DATA': data, 'TYPE': 'TEXT'}

      # 2.3.step prepare sample index
      response_data['QUERY_INDEX'] = 0
      # 2.4.step condition
      response_data['WAITING_TIME'] = self._waiting_time_per_sample
      # 2.5.step total sample number
      self._total_samples = len(self._client_response_record[client_id]['ID'])
      # 2.6.step remained num in this session
      total_num = len(self._client_response_record[client_id]['ID']) \
        if self._max_samples_in_session < 0 else self._max_samples_in_session
      response_data['REMAINED_NUM'] = total_num - 1
      logger.info('leave client id %s "START" response'%client_id)
      return response_data

  def _next_click_branch(self, client_query, record_db, to_next_page=False):
    # continue unfinished session
    client_id = client_query['CLIENT_ID']
    if client_id not in self._client_response_record:
      return {}

    logger.info('enter client id %s "NEXT" response' % client_id)
    now_time = time.time()

    # 1.step client response content
    query_index = client_query['QUERY_INDEX']

    if client_query['QUERY_STATUS'] == 'CLOSE_ECHO' or to_next_page:
      logger.info('prepare next page data')
      # enter next sample
      return self._prepare_next_page(client_id, query_index, record_db)

    # make sure only once access
    # assert(self._client_response_record[client_id]['RESPONSE'][query_index] is None)
    
    if client_query['CLIENT_RESPONSE'] is not None:
      try:
        logger.info('record client id %s response'%client_id)
        user_response = client_query['CLIENT_RESPONSE']
        
        # 1.2.step record client response
        self._client_response_record[client_id]['RESPONSE'][query_index] = user_response
        self._client_response_record[client_id]['RESPONSE_TIME'][query_index]['STOP_TIME'] = now_time
  
        query_data_index = self._client_response_record[client_id]['ID'][query_index]
        if query_data_index not in self._sample_finished_count:
          self._sample_finished_count[query_data_index] = {'CLIENT': [], 'COUNT': 0}
        
        if client_id not in self._sample_finished_count[query_data_index]['CLIENT']:
          self._sample_finished_count[query_data_index]['COUNT'] = \
            self._sample_finished_count[query_data_index]['COUNT'] + 1
          self._sample_finished_count[query_data_index]['CLIENT'].append(client_id)
  
        logger.info('finish record client id %s response'%client_id)
      except:
        logger.error('failed to record client id %s response'%client_id)
        self._client_response_record[client_id]['RESPONSE'][query_index] = None
        self._client_response_record[client_id]['RESPONSE_TIME'][query_index]['STOP_TIME'] = -1
        return {}

    # 2.step server response
    # 2.1.step whether over predefined max time in one session
    if self._max_time_in_session > 0:
      # user has been defined max time in session
      if now_time - self._client_response_record[client_id]['START_TIME'] > self._max_time_in_session:
        # return stop flag
        logger.info('client id %s session time is larger than max time'%client_id)
        return self._prepare_stop_page()

    # 2.2.step whether over predefined max samples in one session
    if self._max_samples_in_session > 0:
      # user has been defined max samples in session
      if (query_index + 1) >= self._max_samples_in_session:
        # return stop flag
        logger.info('client id %s max samples is larger than task defined'%client_id)
        return self._prepare_stop_page()

    # 2.3.step whether need skip first some samples (in first some samples, we must return ground truth)
    if self._skip_sample_num > 0:
      # user has been defined skip first some samples
      if query_index < self._skip_sample_num:
        # user would see the ground truth of his last judgement
        logger.info('prepare grouth truth page')
        return self._prepare_ground_truth_page(client_id, query_index, record_db)

    # 2.4.step enter next sample
    logger.info('prepare next page')
    return self._prepare_next_page(client_id, query_index, record_db)

  def _query_is_legal(self, client_query):
    if 'CLIENT_ID' not in client_query:
      logger.error('client query must contain key "CLIENT_ID"')
      return False

    if 'QUERY' not in client_query:
      logger.error('client query must contain key "QUERY"')
      return False

    if client_query['QUERY'] not in ['START', 'NEXT']:
      logger.error('client query must be "START" or "NEXT"')
      return False
    
    if client_query['QUERY'] != 'START':
      if client_query['CLIENT_ID'] in self._client_response_record:
        # 1.step check query index consistent
        if int(client_query['QUERY_INDEX']) != self._client_response_record[client_query['CLIENT_ID']]['QUERY_INDEX']:
          logger.error('client_id %s query index %d not '
                       'consistent with server query index %d'%(client_query['CLIENT_ID'],
                                                                int(client_query['QUERY_INDEX']),
                                                                int(self._client_response_record[client_query['CLIENT_ID']]['QUERY_INDEX'])))
          return False
  
        # 2.step check client session has been finished
        if self._client_response_record[client_query['CLIENT_ID']]['QUERY_INDEX'] == \
                len(self._client_response_record[client_query['CLIENT_ID']]['ID']):
          logger.error('client_id %s session has been finished'%client_query['CLIENT_ID'])
          return False

    return True

  def crowdsource_server(self, record_db):
    # 0.step search idel server port
    idle_server_port = _pick_idle_port(from_port=40000, to_port=50000, check_count=20)
    if idle_server_port is None:
      logger.error('couldnt find idle port for crowdsoure server')
      return False

    # select crowdsource info
    crowdsource_info = {'title': self._crowdsource_title,
                        'type': self._crowdsource_type,
                        'complex': self._complex_degree,
                        'time': self._crowdsource_start_time,
                        'complete': 0.0,
                        'bonus': self._crowdsource_bonum,
                        'estimated_time': self._crowdsource_estimated_time}

    # 1.step launch crowdsource server (independent process)
    request_queue = Queue()
    response_queue = Queue()

    process = multiprocessing.Process(target=crowdsrouce_server_start,
                                      args=(os.getpid(),
                                            self.experiment_id,
                                            self.app_token,
                                            '/'.join(os.path.normpath(self.dump_dir).split('/')[0:-1]),
                                            self.name,
                                            self.client_html_template,
                                            self.client_keywords_template,
                                            idle_server_port,
                                            crowdsource_info,
                                            request_queue,
                                            response_queue))
    process.start()

    # 2.step listening crowdsource server is OK
    waiting_time = 10
    is_connected = False
    now_time = time.time()
    while not is_connected and (time.time() - now_time) < 60 * 5:
      # waiting crowdsource server in 5 minutes
      try:
        res = requests.get('http://127.0.0.1:%d/heartbeat'%idle_server_port)
        heatbeat_content = json.loads(res.content)
        if 'ALIVE' in heatbeat_content:
          is_connected = True
          break
      except:
        time.sleep(waiting_time)

    if not is_connected:
      logger.error('fail to connect local crowdsource server, couldnt start crowdsource crowdsource server')
      return False

    # 3.step listening client query until crowdsource server is finished
    while not self._is_finished:
      client_id = ''
      try:
        # 2.1 step receive request
        client_query = request_queue.get()

        # 2.2 step check query is legal
        if not self._query_is_legal(client_query):
          if self._client_response_record[client_query['CLIENT_ID']]['QUERY_INDEX'] == \
                  len(self._client_response_record[client_query['CLIENT_ID']]['ID']) - 1:
              # send stop page
              response_queue.put(self._prepare_stop_page())
          else:
            # send unknown page
            response_queue.put({})
          continue
          
        # client id
        client_id = client_query['CLIENT_ID']

        # 2.3 step response client query
        # QUERY: 'START', 'NEXT'
        ########################################################
        #############            START           ###############
        ########################################################
        if client_query['QUERY'] == 'START':
          logger.info('response client_id %s %s query'%(client_id, 'START'))
          response = self._start_click_branch(client_query, record_db)
          response_queue.put(response)
          continue

        ########################################################
        #############            NEXT            ###############
        ########################################################
        if client_query['QUERY'] == 'NEXT':
          logger.info('response client_id %s %s query'%(client_id, 'NEXT'))
          response = self._next_click_branch(client_query, record_db)
          response_queue.put(response)
          continue

        logger.error('client_id %s unknow error'%client_id)
        response_queue.put({})
        continue
      except:
        logger.error('client_id %s unknow error'%client_id)
        response_queue.put({})
    
    # save crowdsource client response
    with open(os.path.join(self.dump_dir, 'crowdsource_record.txt'), 'w') as fp:
      fp.write(json.dumps(self._client_response_record))

    # kill crowdsource server
    # suspend 5 minutes, crowdsource abort suddenly
    time.sleep(30)
    # kill crowdsource server
    os.kill(process.pid, signal.SIGTERM)
    return True


# -*- coding: UTF-8 -*-
# @Time    : 17-6-22
# @File    : train.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals

import tarfile
from datetime import datetime
from multiprocessing import Process
from antgo.ant import flags
from antgo.ant.base import *
from antgo.ant.utils import *
from antgo.dataflow.common import *
from antgo.dataflow.recorder import *
from antgo.measures.statistic import *
from antgo.task.task import *
from antgo.utils.netvis import *
from antgo.utils.serialize import *
from antgo.resource.html import *
from antgo.utils.concurrency import *
from antgo.measures.statistic import *
from antgo.measures.repeat_statistic import *
from antgo.measures.deep_analysis import *
import signal
import traceback
if sys.version > '3':
    PY3 = True
else:
    PY3 = False

FLAGS = flags.AntFLAGS


class AntTrain(AntBase):
  def __init__(self, ant_context,
               ant_name,
               ant_data_folder,
               ant_dump_dir,
               ant_token,
               ant_task_config,
               **kwargs):
    super(AntTrain, self).__init__(ant_name, ant_context, ant_token, **kwargs)
    self.ant_data_source = ant_data_folder
    self.ant_dump_dir = ant_dump_dir
    self.ant_context.ant = self
    self.ant_task_config = ant_task_config
    self.skip_training = kwargs.get('skip_training', False)
    self.context.devices = [int(d) for d in kwargs.get('devices', '').split(',') if d != '']

  def error_analysis(self, running_ant_task, running_ant_dataset, task_running_statictic):
    # error analysis
    logger.info('Start error analysis.')
    # task_running_statictic={self.ant_name:
    #                           {'measure':[
    #                             {'statistic': {'name': 'MESR',
    #                                            'value': [{'name': 'MESR', 'value': 0.4, 'type':'SCALAR'}]},
    #                                            'info': [{'id':0,'score':0.8,'category':1},
    #                                                     {'id':1,'score':0.3,'category':1},
    #                                                     {'id':2,'score':0.9,'category':1},
    #                                                     {'id':3,'score':0.5,'category':1},
    #                                                     {'id':4,'score':1.0,'category':1}]},
    #                             {'statistic': {'name': "SE",
    #                                            'value': [{'name': 'SE', 'value': 0.5, 'type': 'SCALAR'}]},
    #                                            'info': [{'id':0,'score':0.4,'category':1},
    #                                                     {'id':1,'score':0.2,'category':1},
    #                                                     {'id':2,'score':0.1,'category':1},
    #                                                     {'id':3,'score':0.5,'category':1},
    #                                                     {'id':4,'score':0.23,'category':1}]}]}}
    
    # stage-1 error statistic analysis
    for measure_result in task_running_statictic[self.ant_name]['measure']:
      if 'info' in measure_result and len(measure_result['info']) > 0:
        measure_name = measure_result['statistic']['name']
        measure_data = measure_result['info']
    
        # independent analysis per category for classification problem
        measure_data_list = []
        if running_ant_task.class_label is not None and len(running_ant_task.class_label) > 1:
          if running_ant_task.class_label is not None:
            for cl_i, cl in enumerate(running_ant_task.class_label):
              measure_data_list.append([md for md in measure_data if md['category'] == cl or md['category'] == cl_i])
      
        if len(measure_data_list) == 0:
          measure_data_list.append(measure_data)
    
        for category_id, category_measure_data in enumerate(measure_data_list):
          if len(category_measure_data) == 0:
            continue
      
          if 'analysis' not in task_running_statictic[self.ant_name]:
            task_running_statictic[self.ant_name]['analysis'] = {}
      
          if measure_name not in task_running_statictic[self.ant_name]['analysis']:
            task_running_statictic[self.ant_name]['analysis'][measure_name] = {}
      
          # reorganize as list
          method_samples_list = [{'name': self.ant_name, 'data': category_measure_data}]

          # reorganize data as score matrix
          method_num = len(method_samples_list)
          # samples_num are the same among methods
          samples_num = len(method_samples_list[0]['data'])
          # samples_num = ant_test_dataset.size
          method_measure_mat = np.zeros((method_num, samples_num))
          samples_map = []
      
          for method_id, method_measure_data in enumerate(method_samples_list):
            # reorder data by index
            order_key = 'id'
            if 'index' in method_measure_data['data'][0]:
              order_key = 'index'
            method_measure_data_order = sorted(method_measure_data['data'], key=lambda x: x[order_key])
        
            if method_id == 0:
              # record sample id
              for sample_id, sample in enumerate(method_measure_data_order):
                samples_map.append(sample)
        
            # order consistent
            for sample_id, sample in enumerate(method_measure_data_order):
              method_measure_mat[method_id, sample_id] = sample['score']
      
          is_binary = False
          # collect all score
          test_score = [td['score'] for td in method_samples_list[0]['data']
                        if td['score'] > -float("inf") and td['score'] < float("inf")]
          hist, x_bins = np.histogram(test_score, 100)
          if len(np.where(hist > 0.0)[0]) <= 2:
            is_binary = True
      
          # score matrix analysis
          if not is_binary:
            s, ri, ci, lr_samples, mr_samples, hr_samples = \
              continuous_multi_model_measure_analysis(method_measure_mat, samples_map, running_ant_dataset)
        
            analysis_tag = 'Global'
            if len(measure_data_list) > 1:
              analysis_tag = 'Global-Category-' + str(running_ant_task.class_label[category_id])
        
            model_name_ri = [method_samples_list[r]['name'] for r in ri]
            task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag] = \
              {'value': s.tolist() if type(s) != list else s,
               'type': 'MATRIX',
               'x': ci,
               'y': model_name_ri,
               'sampling': [{'name': 'High Score Region', 'data': hr_samples},
                            {'name': 'Middle Score Region', 'data': mr_samples},
                            {'name': 'Low Score Region', 'data': lr_samples}]}
        
            # group by tag
            tags = getattr(running_ant_dataset, 'tag', None)
            if tags is not None:
              for tag in tags:
                g_s, g_ri, g_ci, g_lr_samples, g_mr_samples, g_hr_samples = \
                  continuous_multi_model_measure_analysis(method_measure_mat,
                    samples_map,
                    running_ant_dataset,
                    filter_tag=tag)
            
                analysis_tag = 'Group'
                if len(measure_data_list) > 1:
                  analysis_tag = 'Group-Category-' + str(running_ant_task.class_label[category_id])
            
                if analysis_tag not in task_running_statictic[self.ant_name]['analysis'][measure_name]:
                  task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag] = []
            
                model_name_ri = [method_samples_list[r]['name'] for r in g_ri]
                tag_data = {'value': g_s.tolist() if type(g_s) != list else g_s,
                            'type': 'MATRIX',
                            'x': g_ci,
                            'y': model_name_ri,
                            'sampling': [{'name': 'High Score Region', 'data': g_hr_samples},
                                         {'name': 'Middle Score Region', 'data': g_mr_samples},
                                         {'name': 'Low Score Region', 'data': g_lr_samples}]}
            
                task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag].append((tag, tag_data))
          else:
            s, ri, ci, region_95, region_52, region_42, region_13, region_one, region_zero = \
              discrete_multi_model_measure_analysis(method_measure_mat,
                samples_map,
                running_ant_dataset)
        
            analysis_tag = 'Global'
            if len(measure_data_list) > 1:
              analysis_tag = 'Global-Category-' + str(running_ant_task.class_label[category_id])
        
            model_name_ri = [method_samples_list[r]['name'] for r in ri]
            task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag] = \
              {'value': s.tolist() if type(s) != list else s,
               'type': 'MATRIX',
               'x': ci,
               'y': model_name_ri,
               'sampling': [{'name': '95%', 'data': region_95},
                            {'name': '52%', 'data': region_52},
                            {'name': '42%', 'data': region_42},
                            {'name': '13%', 'data': region_13},
                            {'name': 'best', 'data': region_one},
                            {'name': 'zero', 'data': region_zero}]}
        
            # group by tag
            tags = getattr(running_ant_dataset, 'tag', None)
            if tags is not None:
              for tag in tags:
                g_s, g_ri, g_ci, g_region_95, g_region_52, g_region_42, g_region_13, g_region_one, g_region_zero = \
                  discrete_multi_model_measure_analysis(method_measure_mat,
                    samples_map,
                    running_ant_dataset,
                    filter_tag=tag)
                # if 'group' not in task_running_statictic[self.ant_name]['analysis'][measure_name]:
                #   task_running_statictic[self.ant_name]['analysis'][measure_name]['group'] = []
                #
                analysis_tag = 'Group'
                if len(measure_data_list) > 1:
                  analysis_tag = 'Group-Category-' + str(running_ant_task.class_label[category_id])
            
                if analysis_tag not in task_running_statictic[self.ant_name]['analysis'][measure_name]:
                  task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag] = []
            
                model_name_ri = [method_samples_list[r]['name'] for r in g_ri]
                tag_data = {'value': g_s.tolist() if type(g_s) != list else g_s,
                            'type': 'MATRIX',
                            'x': g_ci,
                            'y': model_name_ri,
                            'sampling': [{'name': '95%', 'data': region_95},
                                         {'name': '52%', 'data': region_52},
                                         {'name': '42%', 'data': region_42},
                                         {'name': '13%', 'data': region_13},
                                         {'name': 'best', 'data': region_one},
                                         {'name': 'zero', 'data': region_zero}]}
            
                task_running_statictic[self.ant_name]['analysis'][measure_name][analysis_tag].append((tag, tag_data))
    
    # stage-2 error eye analysis (all or subset in wrong samples)
    eye_analysis_set_size = int(getattr(running_ant_task, 'badcase_size', 100))
    focus_category = int(getattr(running_ant_task, 'badcase_category', 1))
    
    for measure_result in task_running_statictic[self.ant_name]['measure']:
      # analyze measure result
      measure_name = None
      measure_data = None
      if 'info' in measure_result and len(measure_result['info']) > 0:
        measure_name = measure_result['statistic']['name']
        measure_data = measure_result['info']
      
      if measure_name is None or measure_data is None:
        continue

      focus_measure_data = [x for x in measure_data if x['category'] == focus_category]
      if len(focus_measure_data) == 0:
        continue

      measure_obj, = running_ant_task.evaluation_measure(measure_name)
      eye_analysis_error = []
      eye_analysis_all = []
      for sample_result_info in focus_measure_data:
        sample_id = sample_result_info['id']
        sample_score = sample_result_info['score']
        sample_category = sample_result_info['category'] if 'category' in sample_result_info else '-'
        eye_analysis_all.append((sample_id, sample_score, sample_category))
      
      if len(eye_analysis_all) == 0:
        continue
      eye_analysis_all_sort = sorted(eye_analysis_all, key=lambda x: x[1])

      eye_analysis_set_size = min(eye_analysis_set_size, len(eye_analysis_all_sort))
      eye_analysis_set = []
      if measure_obj.larger_is_better:
        eye_analysis_set = eye_analysis_all_sort[0:eye_analysis_set_size]
      else:
        eye_analysis_set = eye_analysis_all_sort[-eye_analysis_set_size:]

      for id, score, category in eye_analysis_set:
        data_attribute_info = {}
        data_attribute_info['id'] = '%d-(c %s)'%(id, str(category))
        data_attribute_info['score'] = score
        
        data_attribute_info['tag'] = []
        sample, label = running_ant_dataset.at(int(id))
        if 'tag' in label:
          if type(label['tag']) == list or type(label['tag']) == tuple:
            data_attribute_info['tag'].extend(list(label['tag']))
          else:
            data_attribute_info['tag'].append(label['tag'])

        if type(sample) == np.ndarray:
          if len(sample.shape) == 2:
            image = ((sample - np.min(sample)) / (np.max(sample) - np.min(sample)) * 255).astype(np.uint8)
            data_attribute_info['data_type'] = 'IMAGE'
            data_attribute_info['data'] = base64.b64encode(png_encode(image, True)).decode('utf-8')
          elif len(sample.shape) == 3:
            image = sample.astype(np.uint8)
            data_attribute_info['data_type'] = 'IMAGE'
            data_attribute_info['data'] = base64.b64encode(png_encode(image, True)).decode('utf-8')
          elif type(sample) == str:
            data_attribute_info['data_type'] = 'STRING'
            data_attribute_info['data'] = sample
          else:
            # TODO support multi images(gif)
            pass

        data_attribute_info['category'] = category
        data_attribute_info['category_type'] = 'STRING'
        if running_ant_task.task_type == 'SEGMENTATION':
          label_map = label['segmentation_map']
          label_map = ((label_map - np.min(label_map)) / (np.max(label_map) - np.min(label_map)) * 255).astype(np.uint8)
          data_attribute_info['category'] = base64.b64encode(png_encode(label_map, True)).decode('utf-8')
          data_attribute_info['category_type'] = 'IMAGE'
        eye_analysis_error.append(data_attribute_info)
      
      if 'eye' not in task_running_statictic[self.ant_name]:
          task_running_statictic[self.ant_name]['eye'] = {}

      task_running_statictic[self.ant_name]['eye'][measure_name] = eye_analysis_error
    return task_running_statictic

  def start(self):
    # 1.step 加载训练任务
    running_ant_task = None
    if self.token is not None:
      # 1.1.step load train task
      response = mlogger.getEnv().dashboard.challenge.get(command=type(self).__name__)
      if response['status'] == 'ERROR':
        # invalid token
        logger.error('Couldnt load train task.')
        self.token = None
      elif response['status'] in ['OK', 'SUSPEND']:
        content = response['content']

        # maybe user token or task token
        if 'task' in content:
          # task token
          challenge_task = create_task_from_json(content)
          if challenge_task is None:
            logger.error('Couldnt load train task.')
            exit(-1)
          running_ant_task = challenge_task
      else:
        # unknow error
        logger.error('Unknow error.')
        exit(-1)

    if running_ant_task is None:
      # 1.2.step load custom task
      if self.ant_task_config is not None:
        custom_task = create_task_from_xml(self.ant_task_config, self.context)
        if custom_task is None:
          logger.error('Couldnt load custom task.')
          exit(-1)
        running_ant_task = custom_task

    assert (running_ant_task is not None)

    # 2.step 注册实验
    experiment_uuid = self.context.experiment_uuid

    # 3.step 分析消除配置
    # analyze ablation blocks
    # user custom devices
    apply_devices = getattr(self.context.params, 'devices', [])
    if apply_devices is None:
      apply_devices = []

    # ablation experiment
    ablation_blocks = getattr(self.context.params, 'ablation', None)
    ablation_method = getattr(self.context.params, 'ablation_method', 'regular')
    if ablation_method is None:
      ablation_method = 'regular'
    assert(ablation_method in ['regular', 'inregular', 'accumulate', 'any', 'fixed'])
    ablation_experiments_devices_num = 0
    if ablation_blocks is not None:
      ablation_experiments_devices_num = 0
      if ablation_method in ['regular', 'inregular', 'accumulate']:
        ablation_experiments_devices_num = len(ablation_blocks)
        if ablation_method == 'inregular' and len(ablation_blocks) == 1:
          logger.warn('Only exists one ablation block %s, couldnt set inregular ablation method.' % ablation_blocks[0])
          ablation_method = 'regular'
      elif ablation_method == 'any':
        for i in range(len(ablation_blocks)):
          ablation_experiments_devices_num += len(list(itertools.combinations(ablation_blocks, i + 1)))
      else:
        # fixed ablation method
        ablation_experiments_devices_num = 0

    # 4.step  备份实验基本信息
    # 4.1.step 打包代码，并上传至云端
    self.stage = 'TRAIN'
    # - backup in dump_dir
    main_folder = self.main_folder
    if not os.path.exists(os.path.join(self.ant_dump_dir, experiment_uuid)):
      os.makedirs(os.path.join(self.ant_dump_dir, experiment_uuid))
    goldcoin = os.path.join(self.ant_dump_dir, experiment_uuid, 'code.tar.gz')
    
    if os.path.exists(goldcoin):
      os.remove(goldcoin)

    logger.info('Prepare package model files.')
    tar = tarfile.open(goldcoin, 'w:gz')

    # 排除dump目录，将当前文件夹下所有数据上传
    for root, dirs, files in os.walk(main_folder):
      if os.path.commonprefix([root, self.ant_dump_dir]) == self.ant_dump_dir:
        continue

      rel_root = os.path.relpath(root, main_folder)
      for f in files:
        tar.add(os.path.join(root, f), arcname=os.path.join(rel_root, f))

    tar.close()
    logger.info('Finish package process.')

    # TODO: 在下一个版本支持模型文件上传
    # # 上传模型代码
    # mlogger.getEnv().dashboard.experiment.upload(MODEL=goldcoin, APP_STAGE=self.stage)
    # logger.info('finish upload model files')

    # 4.2.step 更新实验配置
    if self.app_token is not None or self.user_token is not None:
      mlogger.getEnv().dashboard.experiment.patch(experiment_uuid=experiment_uuid,
                                                  experiment_hyper_parameter=json.dumps(self.ant_context.params.content))

    # 5.step 加载训练数据集
    logger.info('Loading train dataset %s.'%running_ant_task.dataset_name)
    ant_train_dataset = running_ant_task.dataset('train',
                                                 os.path.join(self.ant_data_source, running_ant_task.dataset_name),
                                                 running_ant_task.dataset_params)

    # 5.1.step 检查设置候选数据
    if self.context.params.system['candidate']:
      ant_train_dataset = CandidateDataset(ant_train_dataset)

    # 6.step 消除实验
    # if ablation_blocks is not None:
    #   if len(apply_devices) >= ablation_experiments_devices_num + 1:
    #     ablation_experiments = []
    #     if ablation_method != 'fixed':
    #       ablation_experiments_devices = apply_devices[:ablation_experiments_devices_num]
    #       apply_devices = apply_devices[ablation_experiments_devices_num:]
    #
    #       # assign device to every ablation experiment
    #       ablation_experiments = self.start_ablation_train_proc(ant_train_dataset,
    #                                                             running_ant_task,
    #                                                             ablation_blocks,
    #                                                             ablation_method,
    #                                                             experiment_uuid,
    #                                                             ablation_experiments_devices)
    #       # launch all ablation experiments
    #       logger.info('waiting until all ablation experiments finish')
    #       for ablation_experiment in ablation_experiments:
    #         ablation_experiment.start()
    #
    #     if ablation_method == 'fixed':
    #       for b in ablation_blocks:
    #         self.ant_context.deactivate_block(b)
    #
    #     # launch complete model training process
    #     self.context.params.devices = apply_devices
    #     num_clones = getattr(self.context.params, 'num_clones', None)
    #     if num_clones is not None:
    #       self.context.params.num_clones = len(apply_devices)
    #
    #     self.stage = "TRAIN"
    #     train_dump_dir = os.path.join(self.ant_dump_dir, experiment_uuid, 'train')
    #     if not os.path.exists(train_dump_dir):
    #       os.makedirs(train_dump_dir)
    #
    #     if not self.skip_training:
    #       logger.info('start training process with complete model')
    #       with safe_recorder_manager(ant_train_dataset):
    #         ant_train_dataset.reset_state()
    #         self.context.call_training_process(ant_train_dataset, train_dump_dir)
    #       logger.info('stop main training process with complete model')
    #
    #     # start evaluation and error analysis
    #     logger.info('start evaluation process with complete model')
    #     try:
    #       _, validation_dataset = ant_train_dataset.split(split_params={}, split_method='holdout')
    #       data_annotation_branch = DataAnnotationBranch(Node.inputs(validation_dataset))
    #       self.context.recorder = RecorderNode(Node.inputs(data_annotation_branch.output(1)))
    #
    #       intermediate_dump_dir = os.path.join(self.ant_dump_dir, experiment_uuid, 'record')
    #       with safe_recorder_manager(self.context.recorder):
    #         self.context.recorder.dump_dir = intermediate_dump_dir
    #         with performance_statistic_region(self.ant_name):
    #           self.context.call_infer_process(data_annotation_branch.output(0), train_dump_dir)
    #       self.context.recorder.close()
    #
    #       task_running_statictic = get_performance_statistic(self.ant_name)
    #       task_running_statictic = {self.ant_name: task_running_statictic}
    #       task_running_elapsed_time = task_running_statictic[self.ant_name]['time']['elapsed_time']
    #       task_running_statictic[self.ant_name]['time']['elapsed_time_per_sample'] = \
    #         task_running_elapsed_time / float(validation_dataset.size)
    #
    #       evaluation_measure_result = []
    #       with safe_recorder_manager(RecordReader(intermediate_dump_dir)) as record_reader:
    #         for measure in running_ant_task.evaluation_measures:
    #           if not measure.crowdsource:
    #             # evaluation
    #             record_generator = record_reader.iterate_read('predict', 'groundtruth')
    #             result = measure.eva(record_generator, None)
    #             if measure.is_support_rank:
    #               evaluation_measure_result.append(result)
    #
    #         task_running_statictic[self.ant_name]['measure'] = evaluation_measure_result
    #
    #       # error analysis
    #       task_running_statictic = \
    #         self.error_analysis(running_ant_task, validation_dataset, task_running_statictic)
    #
    #       if task_running_statictic is not None and len(task_running_statictic) > 0:
    #         logger.info('generate model evaluation report')
    #         self.stage = 'EVALUATION-HOLDOUT-REPORT'
    #
    #         # 更新实验统计信息
    #         mlogger.getEnv().dashboard.experiment.patch(experiment_data=
    #                                                 json.dumps({'REPORT': task_running_statictic,
    #                                                              'APP_STAGE': self.stage}))
    #         # 生成实验报告
    #         everything_to_html(task_running_statictic,
    #                            os.path.join(self.ant_dump_dir, experiment_uuid))
    #
    #         # 上传实验报告
    #         logger.info('uploading experiment report')
    #         mlogger.getEnv().dashboard.experiment.upload(
    #           REPORT_HTML=os.path.join(self.ant_dump_dir, experiment_uuid, 'statistic-report.html'),
    #           APP_STAGE=self.stage)
    #
    #       logger.info('stop evaluation process with complete model')
    #     except:
    #       logger.warn('could not find val dataset, and finish evaluation and error analysis')
    #
    #     # join (waiting until all experiments stop)
    #     for ablation_experiment in ablation_experiments:
    #       ablation_experiment.join()
    #     logger.info('all ablation experiments complete')
    #     return
    #
    #   logger.warn('couldnt enable ablation experiment until set devices in *.yaml')

    # 7.step 模型训练实验
    with safe_manager(ant_train_dataset):
      # 7.1.step 模型训练及评估
      # 四种主流估计方法，holdout; repeated-holdout; bootstrap; kfold
      if running_ant_task.estimation_procedure is not None and \
              running_ant_task.estimation_procedure.lower() in ["holdout",
                                                                "repeated-holdout",
                                                                "bootstrap",
                                                                "kfold"]:
        logger.info('Start model training and evaluation process.')

        estimation_procedure = running_ant_task.estimation_procedure.lower()
        estimation_procedure_params = running_ant_task.estimation_procedure_params
        evaluation_measures = running_ant_task.evaluation_measures

        if estimation_procedure == 'holdout':
          evaluation_statistic = self._holdout_validation(ant_train_dataset, running_ant_task, experiment_uuid)
          
          if evaluation_statistic is not None and len(evaluation_statistic) > 0:
            logger.info('Generate model evaluation report.')
            self.stage = 'EVALUATION-HOLDOUT-REPORT'

            # 更新实验统计信息
            if self.app_token is not None or self.user_token is not None:
              mlogger.getEnv().dashboard.experiment.patch(
                experiment_data=json.dumps({'REPORT': evaluation_statistic,
                                            'APP_STAGE': self.stage}))

            # 生成实验报告
            everything_to_html(evaluation_statistic,
                               os.path.join(self.ant_dump_dir, experiment_uuid))

            # 上传实验报告
            if self.app_token is not None or self.user_token is not None:
              logger.info('Uploading experiment report.')
              mlogger.getEnv().dashboard.experiment.upload(
                REPORT_HTML=os.path.join(self.ant_dump_dir, experiment_uuid, 'statistic-report.html'),
                APP_STAGE=self.stage)
        elif estimation_procedure == "repeated-holdout":
          number_repeats = 2              # default value
          is_stratified_sampling = True   # default value
          split_ratio = 0.7               # default value (andrew ng, machine learning yearning)
          if estimation_procedure_params is not None:
            number_repeats = int(estimation_procedure_params.get('number_repeats', number_repeats))
            is_stratified_sampling = int(estimation_procedure_params.get('stratified_sampling', is_stratified_sampling))
            split_ratio = float(estimation_procedure_params.get('split_ratio', split_ratio))

          # start model estimation procedure
          evaluation_statistic = self._repeated_holdout_validation(number_repeats,
                                                                   ant_train_dataset,
                                                                   split_ratio,
                                                                   is_stratified_sampling,
                                                                   evaluation_measures,
                                                                   experiment_uuid)
          logger.info('Generate model evaluation report.')
          self.stage = 'EVALUATION-REPEATEDHOLDOUT-REPORT'

          # 更新实验统计信息
          if self.app_token is not None or self.user_token is not None:
            mlogger.getEnv().dashboard.experiment.patch(experiment_data=
                                                        json.dumps({'REPORT': evaluation_statistic,
                                                                    'APP_STAGE': self.stage}))
          # 生成实验报告
          everything_to_html(evaluation_statistic,
                             os.path.join(self.ant_dump_dir, experiment_uuid))

          # 上传实验报告
          if self.app_token is not None or self.user_token is not None:
            logger.info('Uploading experiment report.')
            mlogger.getEnv().dashboard.experiment.upload(
              REPORT_HTML=os.path.join(self.ant_dump_dir, experiment_uuid, 'statistic-report.html'),
              APP_STAGE=self.stage)
        elif estimation_procedure == "bootstrap":
          bootstrap_counts = 5
          if estimation_procedure_params is not None:
            bootstrap_counts = int(estimation_procedure_params.get('bootstrap_counts', bootstrap_counts))
          evaluation_statistic = self._bootstrap_validation(bootstrap_counts,
                                                            ant_train_dataset,
                                                            evaluation_measures,
                                                            experiment_uuid)
          logger.info('Generate model evaluation report.')
          self.stage = 'EVALUATION-BOOTSTRAP-REPORT'

          # 更新实验统计信息
          if self.app_token is not None or self.user_token is not None:
            mlogger.getEnv().dashboard.experiment.patch(experiment_data=
                                                    json.dumps({'REPORT': evaluation_statistic,
                                                                'APP_STAGE': self.stage}))
          # 生成实验报告
          everything_to_html(evaluation_statistic,
                             os.path.join(self.ant_dump_dir, experiment_uuid))

          # 上传实验报告
          if self.app_token is not None or self.user_token is not None:
            logger.info('Uploading experiment report.')
            mlogger.getEnv().dashboard.experiment.upload(
              REPORT_HTML=os.path.join(self.ant_dump_dir, experiment_uuid, 'statistic-report.html'),
              APP_STAGE=self.stage)
        else:
          kfolds = 5
          if estimation_procedure_params is not None:
            kfolds = int(estimation_procedure_params.get('kfold', kfolds))
          evaluation_statistic = self._kfold_cross_validation(kfolds,
                                                              ant_train_dataset,
                                                              evaluation_measures,
                                                              experiment_uuid)

          logger.info('Generate model evaluation report.')
          self.stage = 'EVALUATION-KFOLD-REPORT'

          # 更新实验统计信息
          if self.app_token is not None or self.user_token is not None:
            mlogger.getEnv().dashboard.experiment.patch(experiment_data=
                                                    json.dumps({'REPORT': evaluation_statistic,
                                                                'APP_STAGE': self.stage}))
          # 生成实验报告
          everything_to_html(evaluation_statistic,
                             os.path.join(self.ant_dump_dir, experiment_uuid))

          # 上传实验报告
          if self.app_token is not None or self.user_token is not None:
            logger.info('Uploading experiment report.')
            mlogger.getEnv().dashboard.experiment.upload(
              REPORT_HTML=os.path.join(self.ant_dump_dir, experiment_uuid, 'statistic-report.html'),
              APP_STAGE=self.stage)

        return

      # 7.2.step 模型训练 (忽略 --skip_training)
      self.stage = "TRAIN"
      train_dump_dir = os.path.join(self.ant_dump_dir, experiment_uuid, 'train')
      if not os.path.exists(train_dump_dir):
          os.makedirs(train_dump_dir)

      logger.info('Start training process.')
      ant_train_dataset.reset_state()
      try:
        self.context.call_training_process(ant_train_dataset, train_dump_dir)
      except Exception as e:
        if type(e.__cause__) != StopIteration:
          print(e)
          traceback.print_exc()
      
      self.context.from_experiment = train_dump_dir
      logger.info('Stop training process.')

  def _holdout_validation(self, train_dataset, running_ant_task, now_time):
    # 1.step split train set and validation set
    part_train_dataset, part_validation_dataset = \
      train_dataset.split(split_params={},
                          split_method='holdout')
    part_train_dataset.reset_state()

    # dump_dir
    dump_dir = os.path.join(self.ant_dump_dir, now_time, 'train')
    if not os.path.exists(dump_dir):
      os.makedirs(dump_dir)

    self.stage = 'EVALUATION-HOLDOUT-TRAIN'
    if not self.skip_training:
      # training model
      logger.info("Traning process.")
      try:
        self.context.call_training_process(part_train_dataset, dump_dir)
      except Exception as e:
          if type(e.__cause__) != StopIteration:
            print(e)
            traceback.print_exc()

      self.context.from_experiment = dump_dir
    
    logger.info("Infer process.")
    # data_annotation_branch = DataAnnotationBranch(Node.inputs(part_validation_dataset))
    self.context.recorder = RecorderNode2()
    intermediate_dump_dir = os.path.join(self.ant_dump_dir, dump_dir, 'record')
    
    with safe_recorder_manager(self.context.recorder):
      # self.context.recorder.dump_dir = intermediate_dump_dir
      with performance_statistic_region(self.ant_name):
        try:
          self.context.call_infer_process(part_validation_dataset, dump_dir)
        except Exception as e:
          if type(e.__cause__) != StopIteration:
            print(e)
            traceback.print_exc()

    task_running_statictic = get_performance_statistic(self.ant_name)
    task_running_statictic = {self.ant_name: task_running_statictic}
    task_running_elapsed_time = task_running_statictic[self.ant_name]['time']['elapsed_time']
    task_running_statictic[self.ant_name]['time']['elapsed_time_per_sample'] = \
      task_running_elapsed_time / float(part_validation_dataset.size)

    evaluation_measure_result = []
    no_rank_measure = []
    with safe_recorder_manager(RecordReader(intermediate_dump_dir)) as record_reader:
      for measure in running_ant_task.evaluation_measures:
        if not measure.crowdsource:
          # evaluation
          record_generator = record_reader.iterate_read('predict', 'groundtruth')
          result = measure.eva(record_generator, None)
          if measure.is_support_rank:
            evaluation_measure_result.append(result)
          else:
            no_rank_measure.append(result)
  
      task_running_statictic[self.ant_name]['measure'] = evaluation_measure_result

    ################
    # 临时代码
    # voc_measure = {'statistic': {'name': 'voc',
    #                              'value': [{'name': 'MAP', 'value': [23.0, 11.0, 12.0], 'type': 'SCALAR', 'x': 'class',
    #                                         'y': 'Mean Average Precision'},
    #                                        {'name': 'Mean-MAP', 'value': 0.13, 'type': 'SCALAR'}]}}
    #
    # roc_auc_measure = {'statistic': {'name': 'roc_auc',
    #                                  'value': [{'name': 'ROC', 'value': [[[0, 0], [1, 1], [2, 3]],
    #                                                                      [[0, 3], [2, 5], [3, 0]]],
    #                                             'type': 'CURVE', 'x': 'FP', 'y': 'TP',
    #                                             'legend': ['class-0', 'class-1']},
    #                                            {'name': 'AUC', 'value': [0.3, 0.4], 'type': 'SCALAR', 'x': 'class',
    #                                             'y': 'AUC'}]}}
    #
    # pr_f1_measure = {'statistic': {'name': 'pr_f1',
    #                                'value': [{'name': 'Precision-Recall',
    #                                           'value': [[[0, 0], [1, 1], [2, 3]],
    #                                                     [[0, 3], [2, 5], [3, 0]]],
    #                                           'type': 'CURVE', 'x': 'precision', 'y': 'recall'},
    #                                          {'name': 'F1', 'value': [1.0, 2.0], 'type': 'SCALAR', 'x': 'class',
    #                                           'y': 'F1'}]}}
    #
    # confusion_m = {'statistic': {'name': 'cm',
    #                              'value': [{'name': 'ccmm', 'value': (np.ones((3, 4)) * 3).tolist(), 'type': 'MATRIX',
    #                                         'x': ['a', 'b', 'c', 'd'], 'y': ['x', 'y', 'z']}]}}
    #
    # random_img = np.random.random((100, 100))
    # random_img = random_img * 255
    # random_img = random_img.astype(np.uint8)
    # image_m = {'statistic': {'name': 'image',
    #                          'value': [{'name': 'image', 'value': random_img, 'type': 'IMAGE'}]}}
    # task_running_statictic[self.ant_name]['measure'].extend([voc_measure, roc_auc_measure,pr_f1_measure, confusion_m, image_m])
    #
    # task_running_statictic[self.ant_name]['significant_diff'] = {}
    # task_running_statictic[self.ant_name]['significant_diff']['QAS'] = [{'name': '20180429.161714.105706', 'score': 0},
    #                                                                    {'name': '20180429.220804.516647', 'score': 1},
    #                                                                    {'name': '20180429.222627.393376', 'score': -1},
    #                                                                    {'name': '20180429.222721.264422', 'score': 1}]
    # ################

    # error analysis
    task_running_statictic = self.error_analysis(running_ant_task, part_validation_dataset, task_running_statictic)
    task_running_statictic[self.ant_name]['measure'].extend(no_rank_measure)
    return task_running_statictic

  def _repeated_holdout_validation(self,
                                   repeats,
                                   train_dataset,
                                   split_ratio,
                                   is_stratified_sampling,
                                   evaluation_measures,
                                   nowtime):
    
    from_experiment = getattr(self.context, 'from_experiment', None)
    if from_experiment is not None:
      self.context.from_experiment = None
      
    repeated_running_statistic = []
    for repeat in range(repeats):
      # 1.step split train set and validation set
      part_train_dataset, part_validation_dataset = \
        train_dataset.split(split_params={'ratio': split_ratio,
                                          'is_stratified': is_stratified_sampling},
                            split_method='repeated-holdout')
      part_train_dataset.reset_state()

      # dump_dir
      dump_dir = os.path.join(self.ant_dump_dir, nowtime, 'train', 'repeated-holdout-evaluation', 'repeat-%d'%repeat)
      if not os.path.exists(dump_dir):
        os.makedirs(dump_dir)

      # 2.step training model
      self.stage = 'EVALUATION-REPEATEDHOLDOUT-TRAIN-%d' % repeat
      self.context.from_experiment = from_experiment
      logger.info('Start training process at repeathold %d round.'%repeat)
      try:
        self.context.call_training_process(part_train_dataset, dump_dir)
      except Exception as e:
          if type(e.__cause__) != StopIteration:
            print(e)
            traceback.print_exc()

      self.context.from_experiment = dump_dir
      logger.info('Stop training process at repeathold %d round.' % repeat)
      
      # 3.step evaluation measures
      # split data and label
      logger.info('Start infer process at repeathold %d round.' % repeat)
      data_annotation_branch = DataAnnotationBranch(Node.inputs(part_validation_dataset))
      self.context.recorder = RecorderNode(Node.inputs(data_annotation_branch.output(1)))
      
      self.context.from_experiment = None
      self.stage = 'EVALUATION-REPEATEDHOLDOUT-EVALUATION-%d' % repeat
      with safe_recorder_manager(self.context.recorder):
        with performance_statistic_region(self.ant_name):
          try:
            self.context.call_infer_process(data_annotation_branch.output(0), dump_dir)
          except Exception as e:
            if type(e.__cause__) != StopIteration:
              print(e)
              traceback.print_exc()

      logger.info('Start infer process at repeathold %d round.' % repeat)
      
      # clear
      self.context.recorder = None

      task_running_statictic = get_performance_statistic(self.ant_name)
      task_running_statictic = {self.ant_name: task_running_statictic}
      task_running_elapsed_time = task_running_statictic[self.ant_name]['time']['elapsed_time']
      task_running_statictic[self.ant_name]['time']['elapsed_time_per_sample'] = \
          task_running_elapsed_time / float(part_validation_dataset.size)

      logger.info('Start evaluation process at repeathold %d round.'%repeat)
      evaluation_measure_result = []

      with safe_recorder_manager(RecordReader(dump_dir)) as record_reader:
        for measure in evaluation_measures:
          record_generator = record_reader.iterate_read('predict', 'groundtruth')
          result = measure.eva(record_generator, None)
          evaluation_measure_result.append(result)
        task_running_statictic[self.ant_name]['measure'] = evaluation_measure_result
      
      logger.info('Stop evaluation process at repeathold %d round.'%repeat)
      repeated_running_statistic.append(task_running_statictic)

    evaluation_result = \
      multi_repeats_measures_statistic(repeated_running_statistic,
                                       method='repeated-holdout')
    return evaluation_result

  def _bootstrap_validation(self, bootstrap_rounds, train_dataset, evaluation_measures, nowtime):
    bootstrap_running_statistic = []
    from_experiment = getattr(self.context, 'from_experiment', None)
    if from_experiment is not None:
      self.context.from_experiment = None
      
    for bootstrap_i in range(bootstrap_rounds):
      # 1.step split train set and validation set
      part_train_dataset, part_validation_dataset = \
        train_dataset.split(split_params={},
                            split_method='bootstrap')

      # dump_dir
      dump_dir = os.path.join(self.ant_dump_dir,
                              nowtime,
                              'train',
                              'bootstrap-evaluation',
                              'bootstrap-%d-evaluation' % bootstrap_i)
      if not os.path.exists(dump_dir):
        os.makedirs(dump_dir)

      # 2.step training model
      self.stage = 'EVALUATION-BOOTSTRAP-TRAIN-%d' % bootstrap_i
      self.context.from_experiment = from_experiment
      logger.info('Start training process at bootstrap %d round.'%bootstrap_i)
      try:
        self.context.call_training_process(part_train_dataset, dump_dir)
      except Exception as e:
          if type(e.__cause__) != StopIteration:
            print(e)
            traceback.print_exc()

      self.context.from_experiment = dump_dir
      logger.info('Stop training process at bootstrap %d round.' % bootstrap_i)
      
      # 3.step evaluation measures
      # split data and label
      logger.info('Start infer process at bootstrap %d round.' % bootstrap_i)
      data_annotation_branch = DataAnnotationBranch(Node.inputs(part_validation_dataset))
      self.context.recorder = RecorderNode(Node.inputs(data_annotation_branch.output(1)))
      self.context.from_experiment = None
      self.stage = 'EVALUATION-BOOTSTRAP-EVALUATION-%d' % bootstrap_i
      with safe_recorder_manager(self.context.recorder):
        with performance_statistic_region(self.ant_name):
          try:
            self.context.call_infer_process(data_annotation_branch.output(0), dump_dir)
          except Exception as e:
            if type(e.__cause__) != StopIteration:
              print(e)
              traceback.print_exc()

      logger.info('Stop infer process at bootstrap %d round.' % bootstrap_i)
      
      # clear
      self.context.recorder = None

      task_running_statictic = get_performance_statistic(self.ant_name)
      task_running_statictic = {self.ant_name: task_running_statictic}
      task_running_elapsed_time = task_running_statictic[self.ant_name]['time']['elapsed_time']
      task_running_statictic[self.ant_name]['time']['elapsed_time_per_sample'] = \
          task_running_elapsed_time / float(part_validation_dataset.size)

      logger.info('Start evaluation process at bootstrap %d round.' % bootstrap_i)
      evaluation_measure_result = []

      with safe_recorder_manager(RecordReader(dump_dir)) as record_reader:
        for measure in evaluation_measures:
          record_generator = record_reader.iterate_read('predict', 'groundtruth')
          result = measure.eva(record_generator, None)
          evaluation_measure_result.append(result)
        task_running_statictic[self.ant_name]['measure'] = evaluation_measure_result
      
      logger.info('Stop evaluation process at bootstrap %d round.' % bootstrap_i)
      
      bootstrap_running_statistic.append(task_running_statictic)

    evaluation_result = \
      multi_repeats_measures_statistic(bootstrap_running_statistic,
                                       method='bootstrap')
    return evaluation_result

  def _kfold_cross_validation(self, kfolds, train_dataset, evaluation_measures, nowtime):
    # assert (kfolds in [5, 10])
    kfolds_running_statistic = []
    from_experiment = getattr(self.context, 'from_experiment', None)
    if from_experiment is not None:
      self.context.from_experiment = None

    for k in range(kfolds):
      # 1.step split train set and validation set
      part_train_dataset, part_validation_dataset = \
        train_dataset.split(split_params={'kfold': kfolds,
                                          'k': k},
                            split_method='kfold')

      part_train_dataset.reset_state()
      # dump_dir
      dump_dir = os.path.join(self.ant_dump_dir, nowtime, 'train', 'kfold-evaluation', 'fold-%d-evaluation' % k)
      if not os.path.exists(dump_dir):
        os.makedirs(dump_dir)

      # 2.step training model
      self.stage = 'EVALUATION-KFOLD-TRAIN-%d' % k
      logger.info('Start training process at kfold %d round.'%k)
      self.context.from_experiment = from_experiment
      try:
        self.context.call_training_process(part_train_dataset, dump_dir)
      except Exception as e:
          if type(e.__cause__) != StopIteration:
            print(e)
            traceback.print_exc()

      self.context.from_experiment = dump_dir
      logger.info('Stop training process at kfold %d round.'%k)

      # 3.step evaluation measures
      # split data and label
      logger.info('Start infer process at kfold %d round.'%k)
      data_annotation_branch = DataAnnotationBranch(Node.inputs(part_validation_dataset))
      self.context.recorder = RecorderNode(Node.inputs(data_annotation_branch.output(1)))
      self.context.from_experiment = None
      
      self.stage = 'EVALUATION-KFOLD-EVALUATION-%d' % k
      with safe_recorder_manager(self.context.recorder):
        with performance_statistic_region(self.ant_name):
          try:
            self.context.call_infer_process(data_annotation_branch.output(0), dump_dir)
          except Exception as e:
            if type(e.__cause__) != StopIteration:
              print(e)
              traceback.print_exc()

      logger.info('Stop infer process at kfold %d round.'%k)
      # clear
      self.context.recorder = None

      task_running_statictic = get_performance_statistic(self.ant_name)
      task_running_statictic = {self.ant_name: task_running_statictic}
      task_running_elapsed_time = task_running_statictic[self.ant_name]['time']['elapsed_time']
      task_running_statictic[self.ant_name]['time']['elapsed_time_per_sample'] = \
          task_running_elapsed_time / float(part_validation_dataset.size)

      logger.info('Start evaluation process at kfold %d round.'%k)
      evaluation_measure_result = []

      with safe_recorder_manager(RecordReader(dump_dir)) as record_reader:
        for measure in evaluation_measures:
          record_generator = record_reader.iterate_read('predict', 'groundtruth')
          result = measure.eva(record_generator, None)
          evaluation_measure_result.append(result)
        task_running_statictic[self.ant_name]['measure'] = evaluation_measure_result
      logger.info('Stop evaluation process at kfold %d round.'%k)
      
      kfolds_running_statistic.append(task_running_statictic)

    evaluation_result = multi_repeats_measures_statistic(kfolds_running_statistic, method='kfold')
    return evaluation_result

  # def start_ablation_train_proc(self,
  #                               data_source,
  #                               challenge_task,
  #                               ablation_blocks,
  #                               ablation_method,
  #                               time_stamp,
  #                               spare_devices=None):
  #   if ablation_method is None:
  #     ablation_method = 'regular'
  #   # check ablation method
  #   assert(ablation_method in ['regular', 'inregular', 'accumulate', 'any'])
  #
  #   # child func
  #   def proc_func(handle,
  #                 experiment_data_source,
  #                 experiment_challenge_task,
  #                 ablation_block,
  #                 root_time_stamp,
  #                 spare_device,
  #                 skip_training):
  #     # 1.step proc_func is running in a independent process (clone running environment)
  #     handle.reset()
  #     # reassign running device
  #     handle.context.params.devices = [spare_device]
  #     # only one clone
  #     handle.context.params.num_clones = 1
  #
  #     part_train_dataset = experiment_data_source
  #     part_validation_dataset = None
  #     try:
  #       part_train_dataset, part_validation_dataset = \
  #                                       experiment_data_source.split(split_params={}, split_method='holdout')
  #     except:
  #       logger.warn('could not find val dataset for ablation experiment')
  #
  #     # shuffle
  #     part_train_dataset.reset_state()
  #
  #     # deactivate ablation_block
  #     if type(ablation_block) == list or type(ablation_block) == tuple:
  #       for bb in ablation_block:
  #         handle.context.deactivate_block(bb)
  #       ablation_block = '_'.join(ablation_block)
  #     else:
  #       handle.context.deactivate_block(ablation_block)
  #
  #     logger.info('start ablation experiment %s on device %s' % (ablation_block, str(spare_device)))
  #
  #     # dump_dir for ablation experiment
  #     ablation_dump_dir = os.path.join(handle.ant_dump_dir, root_time_stamp, 'train', 'ablation', ablation_block)
  #     if not os.path.exists(ablation_dump_dir):
  #       os.makedirs(ablation_dump_dir)
  #
  #     # 2.step start training process
  #     handle.stage = 'ABLATION-%s-TRAIN' % ablation_block
  #     logger.info('on {}'.format(handle.stage))
  #     if not skip_training:
  #       handle.context.call_training_process(part_train_dataset, ablation_dump_dir)
  #
  #     # 3.step start evaluation process
  #     if part_validation_dataset is not None:
  #       # split data and label
  #       data_annotation_branch = DataAnnotationBranch(Node.inputs(part_validation_dataset))
  #       handle.context.recorder = RecorderNode(Node.inputs(data_annotation_branch.output(1)))
  #
  #       handle.stage = 'ABLATION-%s-EVALUATION' % ablation_block
  #       logger.info('on {}'.format(handle.stage))
  #       with safe_recorder_manager(handle.context.recorder):
  #         handle.context.call_infer_process(data_annotation_branch.output(0), ablation_dump_dir)
  #       handle.context.recorder.close()
  #
  #       # clear
  #       handle.context.recorder = None
  #
  #       ablation_running_statictic = {handle.ant_name: {}}
  #       ablation_evaluation_measure_result = []
  #
  #       logger.info('compute measure')
  #       with safe_recorder_manager(RecordReader(ablation_dump_dir)) as record_reader:
  #         for measure in experiment_challenge_task.evaluation_measures:
  #           record_generator = record_reader.iterate_read('predict', 'groundtruth')
  #           result = measure.eva(record_generator, None)
  #           ablation_evaluation_measure_result.append(result)
  #
  #       ablation_running_statictic[handle.ant_name]['measure'] = ablation_evaluation_measure_result
  #       handle.stage = 'ABLATION-%s-REPORT' % ablation_block
  #
  #       logger.info('on {}'.format(handle.stage))
  #       # 更新实验统计信息
  #       logger.info('update experiment report')
  #       mlogger.getEnv().dashboard.experiment.patch(experiment_data=
  #                                                 json.dumps({'REPORT': ablation_running_statictic,
  #                                                              'APP_STAGE': handle.stage}))
  #
  #       # 生成消除实验报告
  #       everything_to_html(ablation_running_statictic, ablation_dump_dir)
  #
  #       # 上传实验报告
  #       logger.info('uploading experiment report')
  #       mlogger.getEnv().dashboard.experiment.upload(
  #         REPORT_HTML=os.path.join(ablation_dump_dir, 'statistic-report.html'),
  #         APP_STAGE=handle.stage)
  #
  #     handle.context.wait_until_clear()
  #
  #   if ablation_method is None:
  #     ablation_method = 'regular'
  #
  #   traverse_ablation_blocks = []
  #   if ablation_method == 'regular':
  #     traverse_ablation_blocks.extend(ablation_blocks)
  #   elif ablation_method == 'inregular':
  #     for b in ablation_blocks:
  #       traverse_ablation_blocks.append([m for m in ablation_blocks if m != b])
  #   elif ablation_method == 'accumulate':
  #     accumulate_blocks = []
  #     for block in ablation_blocks:
  #       accumulate_blocks.append(block)
  #       traverse_ablation_blocks.append(copy.deepcopy(accumulate_blocks))
  #   else:
  #     for i in range(len(ablation_blocks)):
  #       aa = list(itertools.combinations(ablation_blocks, i+1))
  #       traverse_ablation_blocks.extend(aa)
  #
  #     traverse_ablation_blocks = [list(m) for m in traverse_ablation_blocks]
  #
  #   ablation_experiments = []
  #   for try_i, try_ablation_blocks in enumerate(traverse_ablation_blocks):
  #     # apply independent process
  #     if type(try_ablation_blocks) != list and type(try_ablation_blocks) != tuple:
  #       try_ablation_blocks = [try_ablation_blocks]
  #
  #     try_ablation_blocks = sorted(try_ablation_blocks)
  #
  #     # launch ablation experiment process
  #     block_ablation_process = Process(target=proc_func,
  #                                      args=(self,
  #                                            data_source,
  #                                            challenge_task,
  #                                            copy.deepcopy(try_ablation_blocks),
  #                                            time_stamp,
  #                                            spare_devices[try_i],
  #                                            self.skip_training),
  #                                      name='%s_ablation_block_%s'%(self.ant_name, '_'.join(copy.deepcopy(try_ablation_blocks))))
  #     ablation_experiments.append(block_ablation_process)
  #
  #   return ablation_experiments


# -*- coding: UTF-8 -*-
# @Time    : 17-6-7
# @File    : common.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import copy
import numpy as np
import sys
from antgo.dataflow.core import *
from antgo.utils import get_rng
from antgo.utils.concurrency import *
try:
    import queue
except:
    import Queue as queue

class BatchData(Node):
  def __init__(self, inputs, batch_size, remainder=False):
    super(BatchData, self).__init__(name=None, action=None, inputs=inputs)
    self.batch_size = batch_size
    self.remainder = remainder
    self.stop_iteration = False
    self.buffer = None

  def _fetch_batch_data(self):
    try:
      if self.stop_iteration:
        raise StopIteration
    
      batch_list = []
      label_list = []
    
      for _ in range(self.batch_size):
        input = self._positional_inputs[0]
        data = input.get_value()
        a = None
        b = None
        if type(data) == tuple or type(data) == list:
          a = data[0]
          b = data[1] if len(data) == 2 else data[1:]
        else:
          a = data
          b = None
        
        batch_list.append(a)
        label_list.append(b)
        if _ != self.batch_size - 1:
          input._force_inputs_dirty()
    
      batch = BatchData._aggregate_batch(batch_list)
      # clear reset state
      self._iteration_reset_state = False
      
      return batch, label_list
    except StopIteration:
      if self.stop_iteration or not self.remainder or len(batch_list) == 0:
        # reset all input
        self._reset_iteration_state()
        self.stop_iteration = False
        raise StopIteration
    
      if self.remainder:
        self.stop_iteration = True
        batch = BatchData._aggregate_batch(batch_list)
        return batch, label_list
    except:
      info = sys.exc_info()
      logger.error('%s:%s' % (info[0], info[1]))
      exit(-1)
  
  def _evaluate(self):
    return self._fetch_batch_data()
    
  @staticmethod
  def _aggregate_batch(batch_list):
    batch = None
    max_shape = np.array([data.shape for data in batch_list]).max(axis=0)
    if len(max_shape) == 1:
      batch = np.zeros((len(batch_list), max_shape[0]), dtype=batch_list[0].dtype)
    elif len(max_shape) == 2:
      batch = np.zeros((len(batch_list), max_shape[0], max_shape[1]), dtype=batch_list[0].dtype)
    elif len(max_shape) == 3:
      batch = np.zeros((len(batch_list), max_shape[0], max_shape[1], max_shape[2]), dtype=batch_list[0].dtype)
    elif len(max_shape) == 4:
      batch = np.zeros((len(batch_list), max_shape[0], max_shape[1], max_shape[2], max_shape[3]),
          dtype=batch_list[0].dtype)

    assert (batch is not None)

    for i in range(len(batch_list)):
      batch[i] = batch_list[i]

    return batch


class RandomChooseData(Node):
  def __init__(self, inputs, propabilitys):
    super(RandomChooseData, self).__init__(name=None, action=None, inputs=inputs)
    self.input_list = []
    self.input_list.extend([i for i in self._positional_inputs])

    self.propability_list = tuple(propabilitys)
    self.rng = get_rng(self)

  def _evaluate(self):
    try:
      itr = self.rng.choice(self.input_list, p=self.propability_list)
      # clear reset state
      self._iteration_reset_state = False
      return itr.get_value()
    except StopIteration:
      # reset all input
      self._reset_iteration_state()
      raise StopIteration
    except:
      info = sys.exc_info()
      logger.error('%s:%s'%(info[0],info[1]))
      exit(-1)


class RandomMixData(Node):
  def __init__(self, inputs):
    super(RandomMixData, self).__init__(name=None, action=None, inputs=inputs)
    self.rng = get_rng(self)
    self.input_list_flag = np.ones((len(self._positional_inputs)), dtype=np.uint8)

  def _evaluate(self):
    selected_i = -1
    try:
      ok_inputs = np.where(self.input_list_flag == 1)[0]
      if len(ok_inputs) == 0:
          raise StopIteration

      selected_i = self.rng.randint(low=0, high=len(ok_inputs), size=1)[0]
      selected_i = ok_inputs[selected_i]
      # clear reset state
      self._iteration_reset_state = False
      return self._positional_inputs[selected_i].get_value()
    except StopIteration:
      if selected_i >= 0:
        self.input_list_flag[selected_i] = 0

        while True:
          # select the next valid data input
          try_selected_i = -1
          try:
            maybe_ok_inputs = np.where(self.input_list_flag == 1)[0]
            if len(maybe_ok_inputs) == 0:
                raise StopIteration

            try_selected_i = self.rng.randint(low=0, high=len(maybe_ok_inputs), size=1)[0]
            try_selected_i = maybe_ok_inputs[try_selected_i]
            return self._positional_inputs[try_selected_i].get_value()
          except StopIteration:
            if try_selected_i >= 0:
                self.input_list_flag[try_selected_i] = 0
            else:
                break
          except:
            pass

      raise StopIteration
    except:
      info = sys.exc_info()
      logger.error('%s:%s'%(info[0],info[1]))
      exit(-1)


class JoinData(Node):
  def __init__(self, inputs):
    super(JoinData, self).__init__(name=None, action=None, inputs=inputs)
    self.index = 0

  def _evaluate(self):
    try:
      try:
        self._iteration_reset_state = False
        return self._positional_inputs[self.index].get_value()
      except StopIteration:
        self.index = (self.index + 1) % len(self._positional_inputs)
        if self.index == 0:
          raise StopIteration
        self._iteration_reset_state = False
        return self._positional_inputs[self.index].get_value()
    except StopIteration:
      # reset all input
      self._reset_iteration_state()
      raise StopIteration
    except:
      info = sys.exc_info()
      logger.error('%s:%s'%(info[0],info[1]))
      exit(-1)


class SerializedData(Node):
  def __init__(self, inputs):
    super(SerializedData, self).__init__(name=None, action=None, inputs=inputs)
    self._data_generator = None
    self._buffer_is_empty = False

  def _buffer_pool(self):
    all_values = []
    for i in self._positional_inputs:
      value = i.get_value()
      if type(value) != list:
        value = [value]

      all_values.extend(value)

    for name, i in items(self._keyword_inputs):
      value = i.get_value()
      if type(value) != list:
          value = [value]

      all_values.extend(value)

    self._buffer_is_empty = False
    for index in range(len(all_values)):
      yield all_values[index]

      if index == len(all_values) - 1:
        self._buffer_is_empty = True

  def _evaluate(self):
    try:
      if self._data_generator is None:
        self._data_generator = self._buffer_pool()
        self._buffer_is_empty = False

      # clear reset state
      self._iteration_reset_state = False

      return next(self._data_generator)
    except StopIteration:
      if self._buffer_is_empty:
        # force all inputs dirty
        for i in self._positional_inputs:
          i._force_inputs_dirty()
        for name, i in items(self._keyword_inputs):
          i._force_inputs_dirty()

        self._data_generator = self._buffer_pool()
        self._buffer_is_empty = False
        return next(self._data_generator)

      # reset all input
      self._reset_iteration_state()
      raise StopIteration
    except:
      info = sys.exc_info()
      logger.error('%s:%s'%(info[0], info[1]))
      exit(-1)

  def _force_inputs_dirty(self):
    self._value = DIRTY


class _TransparantNode(Node):
  def __init__(self, upper_node, name, active_request=True):
    super(_TransparantNode, self).__init__(name=name, inputs=upper_node)
    assert(len(self._positional_inputs) == 1)
    self.active_request = active_request
    self._buffer = queue.Queue()

    self._model_val = None

  def set_value(self, new_value):
    self._buffer.put(new_value)
    self._set_value(new_value)

  def get_value(self):
    if DIRTY == self._value:
      self._evaluate()

    if self._buffer.qsize() == 0:
      return self._value

    return self._buffer.get()

  def _evaluate(self):
    try:
      self._positional_inputs[0]._evaluate()

      # clear reset state
      self._iteration_reset_state = False
    except StopIteration:
      # reset all input
      self._reset_iteration_state()
      raise StopIteration
    except:
      info = sys.exc_info()
      logger.error('%s:%s' % (info[0], info[1]))
      exit(-1)

  def _force_inputs_dirty(self):
    if self.active_request:
      if self._buffer.qsize() > 0:
        return
      
      if DIRTY == self._value:
        return
      
      for i in self._positional_inputs:
        i._force_inputs_dirty()
      for name, i in items(self._keyword_inputs):
        i._force_inputs_dirty()
      self._value = DIRTY

  @property
  def model_fn(self):
    # dataset format - tfrecord
    return self._model_val
  @model_fn.setter
  def model_fn(self, model_val):
    self._model_val = model_val


class DataAnnotationBranch(Node):
  def __init__(self, inputs):
    super(DataAnnotationBranch, self).__init__(name=None, action=self.action, inputs=inputs)
    self.annotation_branch = _TransparantNode(upper_node=Node.inputs(self), name='annotation')
    self._y = None
    for k, v in self._positional_inputs[0].__dict__.items():
      setattr(self, k, v)

  def action(self, *args, **kwargs):
    image, annotation = args[0] if len(args[0]) == 2 else (args[0], {})
    self.annotation_branch.set_value(annotation)
    return image

  def output(self, index=0):
    if index == 0:
      return self
    else:
      return self.annotation_branch
  
  def model_fn(self, *args, **kwargs):
    # dataset format - tfrecord
    x, self._y = self._positional_inputs[0].model_fn(*args, **kwargs)
    self.annotation_branch.model_fn = self._y
    return x

  @property
  def size(self):
    return self._positional_inputs[0].size

  
class AutoBranch(Node):
  def __init__(self, inputs, branch_func):
    super(AutoBranch, self).__init__(name=None, action=self.action, inputs=inputs)
    self.auto_branch = _TransparantNode(upper_node=Node.inputs(self), name='auto_branch_transparant',active_request=False)
    self.branch_func = branch_func
    
  def action(self, *args, **kwargs):
    image, annotation = args[0] if len(args[0]) == 2 else (args[0], {})
    branch_data = self.branch_func(image, annotation)
    self.auto_branch.set_value(branch_data)

    return image, annotation

  def output(self, index=0):
    if index == 0:
      return self
    else:
      return self.auto_branch
    
    
class SplitMultiBranches(Node):
  def __init__(self, inputs, branches_num=2):
    super(SplitMultiBranches, self).__init__(name=None, action=None, inputs=inputs)
    assert(len(self._positional_inputs) == 0)

    self.branches = [_TransparantNode(upper_node=Node.inputs(self), name='%s_branch_%d'%(self.name, _),
                                      active_request=False) for _ in range(branches_num)]

  def _evaluate(self):
    try:
      for branch in self.branches:
        if branch.is_dirty():
          branch.set_value(self._positional_inputs[0].get_value())
          self._positional_inputs[0]._force_inputs_dirty()
    except StopIteration:
      # reset all input
      self._reset_iteration_state()
      raise StopIteration
    except:
      info = sys.exc_info()
      logger.error('%s:%s' % (info[0], info[1]))
      exit(-1)

  def output(self, index=0):
    return self.branches[index]


class Episode(Node):
  def __init__(self, inputs, capacity=100, samples_per_class=1, classes=[]):
      super(Episode, self).__init__(name=None, action=None, inputs=inputs)
      self.samples_per_class = samples_per_class
      self.classes = classes
      self.capacity = capacity
      self.consume_num = 0
      self.episode_train = _TransparantNode(upper_node=Node.inputs(self), name='episode-train', active_request=False)
      self.is_episode_ok = False
      self.episode_classes_map = {c: i for i, c in enumerate(self.classes)}
      self.rng = get_rng(self)

  def _evaluate(self):
    try:
      input = self._positional_inputs[0]
      if not self.is_episode_ok:
        # prepare episode
        episode_train_data = {category_id: [] for category_id in self.classes}
        while not self.is_episode_ok:
          data = input.get_value()
          a, b = data

          if 'bbox' in b or 'segmentation' in b:
            # image with bbox or segmentation datasource
            pair_list = [(b_i, b_c) for b_i,b_c in enumerate(b['category_id'].tolist())]
            self.rng.shuffle(pair_list)
            for b_index, b_category_id in pair_list:
              if b_category_id in self.classes and \
                              len(episode_train_data[b_category_id]) < self.samples_per_class:

                obj_img = None
                obj_mask = None
                obj_bbox = np.zeros((1, 4), dtype=np.float32)
                if 'bbox' in b:
                    x0, y0, x1, y1 = b['bbox'][b_index, :]
                    x0 = np.maximum(int(x0), 0)
                    y0 = np.maximum(int(y0), 0)
                    x1 = np.minimum(int(x1), a.shape[1] - 1)
                    y1 = np.minimum(int(y1), a.shape[0] - 1)
                    obj_img = a[y0:y1, x0:x1, :]
                    obj_bbox[0, 0] = x0
                    obj_bbox[0, 1] = y0
                    obj_bbox[0, 2] = x1
                    obj_bbox[0, 3] = y1

                if 'segmentation' in b:
                    obj_mask_full = b['segmentation'][b_index]
                    yyxx = np.where(obj_mask_full > 0)
                    yy = yyxx[0]
                    xx = yyxx[1]
                    x0 = np.min(xx)
                    y0 = np.min(yy)
                    x1 = np.max(xx)
                    y1 = np.max(yy)
                    obj_img = a[y0:y1, x0:x1, :]
                    obj_mask = obj_mask_full[y0:y1, x0:x1, :]
                    obj_bbox[0, 0] = x0
                    obj_bbox[0, 1] = y0
                    obj_bbox[0, 2] = x1
                    obj_bbox[0, 3] = y1

                annotation = {'bbox': obj_bbox,
                              'category_id': np.array([self.episode_classes_map[b_category_id]]),
                              'category': [b['category'][b_index]],
                              'info': a.shape}
                if obj_mask is not None:
                    annotation['segmentation'] = [obj_mask]

                episode_train_data[b_category_id].append((obj_img, annotation))
                break

            self.is_episode_ok = True
            for k, v in episode_train_data.items():
                if len(v) < self.samples_per_class:
                    self.is_episode_ok = False
                    break
          else:
              # image datasource
              pass

          input._force_inputs_dirty()

          if self.is_episode_ok:
              tt = []
              for k, v in episode_train_data.items():
                  tt.extend(v)

              self.episode_train.set_value(tt)
              self.consume_num = 0

      if self.consume_num >= self.capacity:
          raise StopIteration

      while True:
        a, b = input.get_value()
        if 'bbox' in b or 'segmentation' in b:
          # image with bbox or segmentation datasource
          keep_index = [i for i, l in enumerate(b['category_id'].tolist()) if l in self.classes]

          if len(keep_index) == 0:
              input._force_inputs_dirty()
              continue

          b_cpy = copy.deepcopy(b)
          b_cpy['category'] = [b_cpy['category'][i] for i in keep_index]
          b_cpy['category_id'] = np.array([self.episode_classes_map[b_cpy['category_id'][i]] for i in keep_index])
          if 'supercategory' in b_cpy:
              b_cpy['supercategory'] = [b_cpy['supercategory'][i] for i in keep_index]
          if 'area' in b_cpy:
              b_cpy['area'] = np.array([b_cpy['area'][i] for i in keep_index])
          if 'segmentation' in b_cpy:
              b_cpy['segmentation'] = [b_cpy['segmentation'][i] for i in keep_index]
          if 'bbox' in b_cpy:
              b_cpy['bbox'] = b_cpy['bbox'][keep_index,:]
          if 'iscrowd' in b_cpy:
              b_cpy['iscrowd'] = [b_cpy['iscrowd'][i] for i in keep_index]

          self.consume_num += 1
          return (a, b_cpy)
      else:
        # image datasource
        return (None, None)

    except StopIteration:
      # reset all input
      self._reset_iteration_state()
      self.is_episode_ok = False
      self.consume_num = 0
      raise StopIteration
    except:
      info = sys.exc_info()
      logger.error('%s:%s'%(info[0], info[1]))
      exit(-1)

  def output(self, index=0):
    if index == 0:
      return self
    else:
      return self.episode_train


class Zip(Node):
  def __init__(self, inputs):
    super(Zip, self).__init__(name=None, action=None, inputs=inputs)

  def _evaluate(self):
    try:
      input_num = len(self._positional_inputs)
      zip_data = [None for _ in range(input_num)]
      for n in range(input_num):
          zip_data[n] = self._positional_inputs[n].get_value()
      return zip_data
    except StopIteration:
      # reset all input
      self._reset_iteration_state()
      raise StopIteration
    except:
      info = sys.exc_info()
      logger.error('%s:%s' % (info[0], info[1]))
      exit(-1)


class FilterNode(Node):
  def __init__(self, inputs, filter_condition):
    super(FilterNode, self).__init__(name=None, action=None, inputs=inputs)
    self._filter_condition = filter_condition
  
  def _evaluate(self):
    data = self._positional_inputs[0].get_value()
    while not self._filter_condition(data):
      self._positional_inputs[0]._force_inputs_dirty()
      data = self._positional_inputs[0].get_value()

    return data


class FilterPNNode(Node):
  def __init__(self, inputs, filter_condition):
    super(FilterPNNode, self).__init__(name=None, action=None, inputs=inputs)
    # filter_condition return -1, 0, 1
    self._filter_condition = filter_condition
  
  def _evaluate(self):
    data = self._positional_inputs[0].get_value()
    is_n = self._filter_condition(data)
    while not is_n != -1:
      self._positional_inputs[0]._force_inputs_dirty()
      data = self._positional_inputs[0].get_value()
      is_n = self._filter_condition(data)
    
    return data, is_n


class MultiThreadPipe(Node):
  """
  Same as :class:`MapData`, but start threads to run the mapping function.
  This is useful when the mapping function is the bottleneck, but you don't
  want to start processes for the entire dataflow pipeline.
  Note:
      1. There is tiny communication overhead with threads, but you
         should avoid starting many threads in your main process to reduce GIL contention.
         The threads will only start in the process which calls :meth:`reset_state()`.
         Therefore you can use ``PrefetchDataZMQ(MultiThreadMapData(...), 1)``
         to reduce GIL contention.
      2. Threads run in parallel and can take different time to run the
         mapping function. Therefore the order of datapoints won't be
         preserved, and datapoints from one pass of `df.get_data()` might get
         mixed with datapoints from the next pass.
         You can use **strict mode**, where `MultiThreadMapData.get_data()`
         is guranteed to produce the exact set which `df.get_data()`
         produces. Although the order of data still isn't preserved.
  """

  class _Worker(StoppableThread):
    def __init__(self, inq, outq, evt, map_func):
      super(MultiThreadPipe._Worker, self).__init__(evt)
      self.inq = inq
      self.outq = outq
      self.func = map_func
      self.daemon = True

    def run(self):
      try:
        while True:
          dp = self.queue_get_stoppable(self.inq)
          if self.stopped():
            return

          # cannot ignore None here. will lead to unsynced send/recv
          self.outq.put(self.func(dp))
      except Exception:
        if self.stopped():
          pass  # skip duplicated error messages
        else:
          raise
      finally:
        self.stop()

  def __init__(self, input_node, pipe_nodes, nr_thread, buffer_size=200, strict=False):
    """
    Args:
        input_node (DataFlow): the dataflow to map
        nr_thread (int): number of threads to use
        map_func (callable): datapoint -> datapoint | None
        buffer_size (int): number of datapoints in the buffer
        strict (bool): use "strict mode", see notes above.
    """
    super(MultiThreadPipe, self).__init__(name='pipe')

    self._strict = strict
    self.nr_thread = nr_thread
    self.buffer_size = buffer_size
    self.entry_node = Input(name='entry')
    self.end_node = self.entry_node
    
    for node in pipe_nodes:
      self.end_node = node[0](Node.inputs(self.end_node), **node[1])
    
    def pipe_process_func(x):
      self.entry_node.set_value(x)
      return self.end_node.get_value()

    self._is_start = False
    self._threads = []

    self._input_node = input_node
    self._iter = None
    self._in_queue = queue.Queue()
    self._out_queue = queue.Queue()
    self._evt = threading.Event()
    self._threads = [MultiThreadPipe._Worker(
      self._in_queue, self._out_queue, self._evt, pipe_process_func)
                     for _ in range(self.nr_thread)]

  def _fill_buffer(self):
    n = self.buffer_size - self._in_queue.qsize() - self._out_queue.qsize()
    assert n >= 0, n
    if n == 0:
      return
    try:
      for _ in range(n):
        self._in_queue.put(next(self._iter))
    except StopIteration:
      logger.error("[MultiThreadMapData] buffer_size cannot be larger than the size of the DataFlow!")
      raise

  def _recv(self):
    ret = self._out_queue.get()
    if ret is None:
      assert not self._strict, \
        "[MultiThreadMapData] Map function cannot return None when strict mode is used."
    return ret

  def iterator_value(self):
    if not self._is_start:
      for t in self._threads:
        t.start()
      self._is_start = True

    self._iter = self._input_node.iterator_value()

    # fill bufer
    self._fill_buffer()

    is_over = False
    while True:
      # return value
      yield self._recv()

      if not is_over:
        try:
          self._in_queue.put(next(self._iter))
        except StopIteration:
          is_over = True
      else:
        if self._out_queue.qsize() == 0 and self._in_queue.qsize() == 0:
          self._input_node._reset_iteration_state()
          break

  def set_value(self, new_value):
    raise NotImplementedError

  def get_value(self):
    raise NotImplementedError
  
  # def __del__(self):
  #   if self._evt is not None:
  #     self._evt.set()
  #   for p in self._threads:
  #     p.join()


class ActionDataset(Node):
  def __init__(self, name, func, input):
    super(ActionDataset, self).__init__(name, func, input)
    self.dataset = input[0][0]

  @property
  def size(self):
    return self.dataset.size


#encoding=utf-8
from __future__ import unicode_literals
from __future__ import division

import numpy as np
import unittest

def binary_c_stats(actual,predicated):
    '''
    computing true positive rate, false positive rate,
    true negative rate and false negative rate for binary classification problem
    Parameters
    ----------
    actual: list
            ground truth label
    predicated: list
            predicated label
    Returns
    -------
    TP,FN,TN,FP
    '''
    # transform to array
    if type(actual) == list:
        actual = np.array(actual)
    if type(predicated) == list:
        predicated = np.array(predicated)

    if actual.shape[0] == 0 or predicated.shape[0] == 0:
        return None

    # transform to array(int)
    if actual.dtype != int:
        actual = actual.astype(dtype=int)

    # absorb some error
    if np.max(actual) > 1:
        actual[np.where(actual > 1)[0]] = 1
    if len(predicated.shape) > 1 and predicated.shape[1] > 1:
        predicated = predicated[:,0]
    predicated = np.reshape(predicated,[predicated.size])
    if predicated.dtype != int:
        predicated = predicated.copy()
        predicated[np.where(predicated <= 0.5)[0]] = 0.0
        predicated[np.where(predicated > 0.5)[0]] = 1.0
        predicated = predicated.astype(dtype=int)

    true_index = np.where(actual == 1)[0]
    true_index_num = len(true_index)
    true_positive = len(np.where(predicated[true_index] == 1)[0])
    false_negative = len(true_index) - true_positive

    negative_index = np.where(actual == 0)[0]
    negative_index_num = len(negative_index)
    true_negative = len(np.where(predicated[negative_index] == 0)[0])
    false_positive = len(negative_index) - true_negative

    return true_positive / (true_index_num + 0.000001),\
           false_negative / (true_index_num + 0.000001),\
           true_negative / (negative_index_num + 0.000001),\
           false_positive / (negative_index_num + 0.000001)

def binary_c_stats2(actual, predicated):
    '''
    regular statistics about binary classification task
    Parameters
    ----------
    actual: list
            ground truth label
    predicated: list
            predicated label
    Returns
    -------
    accuracy: (corrected positive + corrected negative) / all samples
    precision: corrected positive / predicated positives
    recall: corrected positive / all positives
    F1-measure:  
    '''
    #transform to array
    if type(actual) == list:
        actual = np.array(actual)
    if type(predicated) == list:
        predicated = np.array(predicated)

    if actual.shape[0] == 0 or predicated.shape[0] == 0:
        return None

    #transform to array(int)
    if actual.dtype != int:
        actual = actual.astype(dtype=int)

    #absorb some error
    if np.max(actual) > 1:
        actual[np.where(actual > 1)[0]] = 1
    if len(predicated.shape) > 1 and predicated.shape[1] > 1:
        predicated = predicated[:,0]
    predicated = np.reshape(predicated,[predicated.size])

    if predicated.dtype != int:
        predicated = predicated.copy()
        predicated[np.where(predicated <= 0.5)[0]] = 0.0
        predicated[np.where(predicated > 0.5)[0]] = 1.0
        predicated = predicated.astype(dtype=int)

    num = actual.shape[0]
    true_index = np.where(actual == 1)[0]
    true_index_num = len(true_index)
    true_positive = len(np.where(predicated[true_index] == 1)[0])
    false_negative = true_index_num - true_positive

    negative_index = np.where(actual == 0)[0]
    negative_index_num = len(negative_index)
    true_negative = len(np.where(predicated[negative_index] == 0)[0])
    false_positive = negative_index_num - true_negative

    accuracy = float(true_positive + true_negative) / (num + 0.000001)
    precision = float(true_positive) / (true_positive + false_positive + 0.000001)
    recall = float(true_positive) / (true_positive + false_negative + 0.000001)
    F1 = float(2.0 * true_positive) / (2.0 * true_positive + false_positive + false_negative + 0.000001)

    return accuracy,precision,recall,F1

class TestBC(unittest.TestCase):
    def test(self):
        a = [1, 0, 0, 1, 1, 1, 0, 1]
        b = [0, 0, 0, 1, 1, 0, 1, 1]

        accuracy, precision, recall, F1 = binary_c_stats2(a, b)
        self.assertAlmostEqual(accuracy,5.0/8.0)
        self.assertAlmostEqual(precision,3.0/4.0)
        self.assertAlmostEqual(recall,3.0/5.0)
        self.assertAlmostEqual(F1,6.0/(2*3.0+1+2))

        TP, FN, TN, FP = binary_c_stats(a, b)
        self.assertAlmostEqual(TP, 3.0 / 5.0)
        self.assertAlmostEqual(FN, 2.0 / 5.0)
        self.assertAlmostEqual(TN, 2.0 / 3.0)
        self.assertAlmostEqual(FP, 1.0 / 3.0)

if __name__ == "__main__":
    unittest.main()


from __future__ import unicode_literals
from __future__ import division

import numpy as np

def capped_log10_likelihood(actual, predicted):
    """
    Computes the capped log10 likelihood, predictions in (0.01, 0.99).
    This function computes the log likelihood between two numbers,
    or for element between a pair of lists or numpy arrays.
    Parameters
    ----------
    actual : int, float, list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double or list of doubles
            The log likelihood error between actual and predicted
    """
    actual = np.array(actual)
    predicted = np.array(predicted)
    predicted[predicted<0.01]=0.01
    predicted[predicted>0.99]=0.99

    score = -(actual*np.log10(predicted)+(1-actual)*np.log10(1-predicted))

    if type(score)==np.ndarray:
        score[np.isnan(score)] = 0
    else:
        if np.isnan(score):
            score = 0
    return score

def capped_binomial_deviance(actual, predicted):
    """
    Computes the capped binomial deviance.
    This function computes the log loss between two lists
    of numbers.
    Parameters
    ----------
    actual : list of numbers, numpy array
             The ground truth value
    predicted : same type as actual
                The predicted value
    Returns
    -------
    score : double
            The log loss between actual and predicted
    """
    return np.mean(capped_log10_likelihood(actual, predicted))


# encoding=utf-8
# @Time    : 17-8-9
# @File    : significance.py
# @Author  :
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.basic import *
import numpy as np
import random


def bootstrap_confidence_interval(data_source, seed, measure, replicas=50):
  num = data_source.size
  random.seed(seed)
  bootstrap_samples = [[random.randint(0, num-1) for _ in range(num)] for _ in range(replicas)]

  result = measure.eva(data_source.iterate_read('predict', 'groundtruth'), None)
  estimated_measure = result['statistic']['value'][0]['value']

  bootstrap_estimated_measures = []
  for bootstrap_sample in bootstrap_samples:
    record_generator = data_source.iterate_sampling_read(bootstrap_sample, 'predict', 'groundtruth')
    result = measure.eva(record_generator, None)

    value = result['statistic']['value'][0]['value']
    bootstrap_estimated_measures.append(value)

  delta_measure = estimated_measure - np.array(bootstrap_estimated_measures)
  sorted_delta_measure = np.sort(delta_measure)

  # percentile method (B.Efron 1981)
  # confidence interval 95% (alpha = 0.05)
  delta_025 = int(replicas * 0.025)
  delta_975 = int(replicas * 0.975)
  
  return (estimated_measure - sorted_delta_measure[delta_975],
          estimated_measure - sorted_delta_measure[delta_025])


def bootstrap_direct_confidence_interval(bootstrap_estimated_measures):
  sorted_measures = np.sort(bootstrap_estimated_measures)

  # percentile method (B.Efron 1981)
  # confidence interval 95% (alpha = 0.05)
  pos_025 = int(len(bootstrap_estimated_measures) * 0.025)
  pos_975 = int(len(bootstrap_estimated_measures) * 0.975)

  measure_025 = sorted_measures[pos_025]
  measure_975 = sorted_measures[pos_975]

  return (float(measure_025), float(measure_975))


def bootstrap_ab_significance_compare(ab_data_source, seed, measure, replicas=50):
  assert(ab_data_source[0].size == ab_data_source[1].size)
  num = ab_data_source[0].size
  random.seed(seed)
  bootstrap_samples = [[random.randint(0, num-1) for _ in range(num)] for _ in range(replicas)]

  a_bootstrap_scores = []
  for bootstrap_sample in bootstrap_samples:
    record_generator = ab_data_source[0].iterate_sampling_read(bootstrap_sample, 'predict', 'groundtruth')
    result = measure.eva(record_generator, None)
    value = result['statistic']['value'][0]['value']
    a_bootstrap_scores.append(value)

  b_bootstrap_scores = []
  for bootstrap_sample in bootstrap_samples:
    record_generator = ab_data_source[1].iterate_sampling_read(bootstrap_sample, 'predict', 'groundtruth')
    result = measure.eva(record_generator, None)
    value = result['statistic']['value'][0]['value']
    b_bootstrap_scores.append(value)

  diff_bootstrap_scores = np.array(a_bootstrap_scores) - np.array(b_bootstrap_scores)
  sorted_diff_scores = np.sort(diff_bootstrap_scores)

  # percentile method (B.Efron 1981)
  # confidence interval 95% (alpha = 0.05)
  pos_025 = int(replicas * 0.025)
  pos_975 = int(replicas * 0.975)

  if getattr(measure, 'larger_is_better', 0) == 1:
    if sorted_diff_scores[pos_025] > 0.0:
      # a is better
      return 1

    if sorted_diff_scores[pos_975] < 0.0:
      # b is better
      return -1
  else:
    if sorted_diff_scores[pos_025] > 0.0:
      # b is better
      return -1

    if sorted_diff_scores[pos_975] < 0.0:
      # a is better
      return 1

  # a,b has no significant difference
  return 0




# -*- coding: UTF-8 -*-
# @Time    : 17-12-27
# @File    : lfw.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset import *
import os
import numpy as np

__all__ = ['LFW']

LFW_URL = 'http://vis-www.cs.umass.edu/lfw/index.html#download'
class LFW(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(LFW, self).__init__(train_or_test, dir, params)
    assert(train_or_test in ['train', 'test','sample'])

    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    self.image_flag = getattr(self, 'image', 'align')
    # 0.step maybe download
    if self.image_flag == 'align':
      if not os.path.exists(os.path.join(self.dir, 'lfw-deepfunneled')):
        self.download(self.dir,
                      file_names=[],
                      default_url='http://vis-www.cs.umass.edu/lfw/lfw-deepfunneled.tgz',
                      auto_untar=True,
                      is_gz=True)
    else:
      if not os.path.exists(os.path.join(self.dir, 'lfw')):
        self.download(self.dir,
                      file_names=[],
                      default_url='http://vis-www.cs.umass.edu/lfw/lfw.tgz',
                      auto_untar=True,
                      is_gz=True)
    
    # 1.step data folder (wild or align)
    if self.image_flag == 'align':
      self._data_folder = os.path.join(self.dir, 'lfw-deepfunneled')
    else:
      self._data_folder = os.path.join(self.dir, 'lfw')
    
    # 2.step data files
    self._persons_list = []
    self._persons_id_str = []
    self._persons_id = []
    for person_id_str in os.listdir(self._data_folder):
      if person_id_str[0] == '.':
        continue
      for person_image in os.listdir(os.path.join(self._data_folder, person_id_str)):
        self._persons_list.append(os.path.join(self._data_folder, person_id_str, person_image))
        self._persons_id_str.append(person_id_str)
    
    id_set = set(self._persons_id_str)
    self.person_id_map = {}
    for s_i, s in enumerate(id_set):
      self.person_id_map[s] = s_i
    
    for person_id_str in self._persons_id_str:
      person_id = self.person_id_map[person_id_str]
      self._persons_id.append(person_id)
    
    self.ids = list(range(len(self._persons_list)))

    # fixed seed
    self.seed = time.time()

    self.view = int(getattr(self, 'view', 0))
    self.pairs = []
    if self.view == 1:
      if self.train_or_test == 'train':
        with open(os.path.join(self.dir, 'pairsDevTrain.txt')) as fp:
          same_pair_num = fp.readline()
          same_pair_num = int(same_pair_num.strip())
          for pair_i in range(same_pair_num):
            content = fp.readline()
            person_name, image_a, image_b = content.strip().split('\t')
            person_a_image = '%s/%s_%04d.jpg'%(person_name, person_name, int(image_a))
            person_b_image = '%s/%s_%04d.jpg'%(person_name, person_name, int(image_b))

            self.pairs.append((person_a_image, person_b_image))

          content = fp.readline()
          while content:
            content = content.strip()
            if content == '':
              break
            diff_person_a_name, diff_person_a_index, diff_person_b_name, diff_person_b_index = \
              content.split('\t')

            diff_person_a_image = '%s/%s_%04d.jpg'%(diff_person_a_name, diff_person_a_name, int(diff_person_a_index))
            diff_person_b_image = '%s/%s_%04d.jpg'%(diff_person_b_name, diff_person_b_name, int(diff_person_b_index))
            self.pairs.append((diff_person_a_image, diff_person_b_image))

            content = fp.readline()
      elif self.train_or_test == 'test':
        with open(os.path.join(self.dir, 'pairsDevTest.txt')) as fp:
          same_pair_num = fp.readline()
          same_pair_num = int(same_pair_num.strip())
          for pair_i in range(same_pair_num):
            content = fp.readline()
            person_name, image_a, image_b = content.strip().split('\t')
            person_a_image = '%s/%s_%04d.jpg' % (person_name, person_name, int(image_a))
            person_b_image = '%s/%s_%04d.jpg' % (person_name, person_name, int(image_b))

            self.pairs.append((person_a_image, person_b_image))

          content = fp.readline()
          while content:
            content = content.strip()
            if content == '':
              break
            diff_person_a_name, diff_person_a_index, diff_person_b_name, diff_person_b_index = \
              content.split('\t')

            diff_person_a_image = '%s/%s_%04d.jpg' % (diff_person_a_name, diff_person_a_name, int(diff_person_a_index))
            diff_person_b_image = '%s/%s_%04d.jpg' % (diff_person_b_name, diff_person_b_name, int(diff_person_b_index))
            self.pairs.append((diff_person_a_image, diff_person_b_image))

            content = fp.readline()
    else:
      if os.path.exists(os.path.join(self.dir, 'pairs.txt')):
        pass
      if os.path.exists(os.path.join(self.dir, 'people.txt')):
        pass
    
  @property
  def size(self):
    if self.view == 0:
      return len(self.ids)
    elif self.view == 1:
      return len(self.pairs)
    else:
      return 0
  
  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1
    
      if self.view == 0:
        idxs = copy.deepcopy(self.ids)
        if self.rng:
          self.rng.shuffle(idxs)

        for k in idxs:
          person_file = self._persons_list[k]
          person_image = imread(person_file)
          person_id_str = self._persons_id_str[k]
          person_id = self._persons_id[k]

          yield [person_image, {'category_id': person_id,
                                'category': person_id_str,
                                'id': k,
                                'info': [person_image.shape[0], person_image.shape[1], person_image.shape[2]]}]
      elif self.view == 1:
        pair_ids = list(range(len(self.pairs)))
        if self.rng:
          self.rng.shuffle(pair_ids)
        for pair_i in pair_ids:
          person_a, person_b = self.pairs[pair_i]
          person_a_name = person_a.split('/')[0]
          person_b_name = person_b.split('/')[0]
          person_a_path = os.path.join(self._data_folder, person_a)
          person_a_image = imread(person_a_path)
          person_b_path = os.path.join(self._data_folder, person_b)
          person_b_image = imread(person_b_path)

          data = np.stack([person_a_image, person_b_image],0)
          yield data, {'a_category': person_a_name,
                       'a_category_id': self.person_id_map[person_a_name],
                       'b_category': person_b_name,
                       'b_category_id': self.person_id_map[person_b_name],
                       'id': pair_i}
      else:
        pass

  def at(self, id):
    if self.train_or_test == 'sample':
      return self.data_samples[id]

    if self.view == 0:
      person_file = self._persons_list[id]
      person_image = imread(person_file)
      person_id_str = self._persons_id_str[id]
      person_id = self._persons_id[id]

      return person_image, {'category_id': person_id,
                            'category': person_id_str,
                            'id': id,
                            'info': [person_image.shape[0], person_image.shape[1], person_image.shape[2]]}
    elif self.view == 1:
      pair_i = id
      person_a, person_b = self.pairs[pair_i]
      person_a_name = person_a.split('/')[0]
      person_b_name = person_b.split('/')[0]
      person_a_path = os.path.join(self.dir, person_a)
      person_a_image = imread(person_a_path)
      person_b_path = os.path.join(self.dir, person_b)
      person_b_image = imread(person_b_path)

      data = np.stack([person_a_image, person_b_image], 0)
      return data, {'a_category': person_a_name,
                   'a_category_id': self.person_id_map[person_a_name],
                   'b_category': person_b_name,
                   'b_category_id': self.person_id_map[person_b_name],
                   'id': pair_i}
    else:
      pass
  
  def split(self, split_params={}, split_method=''):
    assert (self.train_or_test == 'train')
    assert (split_method in ['repeated-holdout', 'bootstrap', 'kfold'])
    
    category_ids = None
    if split_method == 'kfold':
      np.random.seed(np.int64(self.seed))
      category_ids = [i for i in range(len(self.ids))]
      np.random.shuffle(category_ids)
    else:
      category_ids = [self._persons_id[i] for i in range(len(self.ids))]
    
    train_ids, val_ids = self._split(category_ids, split_params, split_method)
    train_dataset = LFW(self.train_or_test, self.dir, self.ext_params)
    train_dataset.ids = train_ids
    
    val_dataset = LFW(self.train_or_test, self.dir, self.ext_params)
    val_dataset.ids = val_ids
    
    return train_dataset, val_dataset


# encoding=utf-8
# @Time    : 17-7-25
# @File    : matting_task.py
# @Author  :
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import numpy as np
from antgo.task.task import *
from antgo.measures.base import *
from antgo.dataflow.common import *
from antgo.measures.error import *

default = {'AntSADMatting': ('MATTING-SAD', 'MATTING'),
           'AntMSEMatting': ('MATTING-MSE', 'MATTING'),
           'AntGradientMatting': ('MATTING-GRADIENT', 'MATTING')}
class AntSADMatting(AntMeasure):
  def __init__(self, task):
    super(AntSADMatting, self).__init__(task, 'MATTING-SAD')
    assert (task.task_type == 'MATTING')

    self.is_support_rank = True

  def eva(self, data, label):
    if label is not None:
      data = zip(data, label)

    count = 0
    sad = 0.0
    for predict, gt in data:
      assert(len(predict.shape) == 2)
      assert(len(gt.shape) == 2)

      sad += np.sum(np.abs(predict - gt))
      count += 1

    val = sad / count
    return {'statistic':{'name':self.name, 'value':[{'name':self.name, 'value': val, 'type': 'SCALAR'}]}}


class AntMSEMatting(AntMeasure):
  def __init__(self, task):
    super(AntMSEMatting, self).__init__(task, 'MATTING-MSE')
    assert (task.task_type == 'MATTING')

    self.is_support_rank = True

  def eva(self, data, label):
    if label is not None:
      data = zip(data, label)

    count = 0
    res = 0.0
    for predict, gt in data:
      assert(len(predict.shape) == 2)
      assert(len(gt.shape) == 2)

      res += mse(gt, predict)
      count += 1

    val = res / count
    return {'statistic': {'name': self.name, 'value': [{'name': self.name, 'value': val, 'type': 'SCALAR'}]}}


class AntBoundaryMSEMatting(AntMeasure):
  def __init__(self, task):
    pass

  def eva(self, data, label):
    pass


class AntGradientMatting(AntMeasure):
  def __init__(self, task):
    # paper: Christoph Rhemann, etc. A Perceptually Motivated Online Benchmark for Image Matting
    super(AntGradientMatting, self).__init__(task, 'MATTING-GRADIENT')
    assert (task.task_type == 'MATTING')
    # delta = 1.4, q = 2

    self.is_support_rank = True

  def eva(self, data, label):
    if label is not None:
      data = zip(data, label)

    count = 0
    res = 0.0
    for predict, gt in data:
      assert(len(predict.shape) == 2)
      assert(len(gt.shape) == 2)

      predict_grad = scipy.ndimage.filters.gaussian_filter(predict, 1.4, order=1)
      gt_grad = scipy.ndimage.filters.gaussian_filter(gt, 1.4, order=1)
      res += np.sum(np.power(predict_grad - gt_grad, 2))
      count += 1

    val = res / count
    return {'statistic': {'name': self.name, 'value': [{'name': self.name, 'value': val, 'type': 'SCALAR'}]}}


class AntConnectivityMatting(AntMeasure):
  def __init__(self, task):
    # paper: Christoph Rhemann, etc. A Perceptually Motivated Online Benchmark for Image Matting
    super(AntConnectivityMatting, self).__init__(task, 'MATTING-CONNECTIVITY')
    assert (task.task_type == 'MATTING')
    # theta=0.15, p=1

    self.is_support_rank = True

  def eva(self, data, label):
    if label is not None:
      data = zip(data, label)

    count = 0
    res = 0.0
    for predict, gt in data:
      assert(len(predict.shape) == 2)
      assert(len(gt.shape) == 2)

      count += 1

    val = 0.0
    return {'statistic': {'name': self.name, 'value': [{'name': self.name, 'value': val, 'type': 'SCALAR'}]}}

